from abc import ABCMeta
from collections import defaultdict, Counter
from operator import attrgetter
from typing import Iterable, List, Optional, Set

from six import add_metaclass

import numpy as np

from glypy.utils import uid
from glypy.structure.glycan_composition import HashableGlycanComposition

from ms_deisotope import DeconvolutedPeak
from ms_deisotope.data_source import ProcessedRandomAccessScanSource

from .mass_shift import MassShiftBase, Unmodified
from .utils import ArithmeticMapping


try:
    trapezoid = np.trapz
except AttributeError:
    trapezoid = np.trapezoid

MIN_POINTS_FOR_CHARGE_STATE = 3
intensity_getter = attrgetter("intensity")


class EmptyListException(Exception):
    pass


class DuplicateNodeError(Exception):
    pass


def group_by(ungrouped_list, key_fn=lambda x: x, transform_fn=lambda x: x):
    groups = defaultdict(list)
    for item in ungrouped_list:
        key_value = key_fn(item)
        groups[key_value].append(transform_fn(item))
    return groups


def split_by_charge(peaks):
    return group_by(peaks, lambda x: x.charge)


def count_charge_states(peaks):
    peaks = [j for i in peaks for j in i]
    return len(split_by_charge(peaks))


class SubsequenceMasker(object):
    def __init__(self, target, masker):
        self.target = target
        self.masker = masker

        self.masking_nodes = []

    def masking_nodes_from_masker(self):
        self.masking_nodes = self.masker.nodes.unspool()

    def masking_nodes_from_peaks(self, peaks):
        nodes = []
        for node in self.masker.nodes.unspool_strip_children():
            for peak in node.members:
                if peak in peaks:
                    nodes.append(node)
                    break
        self.masking_nodes = nodes

    def mask(self):
        unmasked_nodes = []
        target_nodes = self.target.nodes.unspool_strip_children()
        for node in target_nodes:
            if node not in self.masking_nodes:
                unmasked_nodes.append(node)

        new = self.target.clone()
        new.clear()
        new.used_as_mass_shift = []
        new.created_at = "mask_subsequence"
        for node in unmasked_nodes:
            new.insert_node(node)
        return new

    @classmethod
    def mask_subsequence(cls, target, masker, peaks=None):
        inst = cls(target, masker)
        if peaks is None:
            inst.masking_nodes_from_masker()
        else:
            inst.masking_nodes_from_peaks(peaks)
        return inst.mask()


mask_subsequence = SubsequenceMasker.mask_subsequence


class _TimeIntervalMethods(object):
    __slots__ = []

    def overlaps_in_time(self, interval):
        self_start_time = self.start_time
        self_end_time = self.end_time
        interval_start_time = interval.start_time
        interval_end_time = interval.end_time
        cond = ((self_start_time <= interval_start_time and self_end_time >= interval_end_time) or (
            self_start_time >= interval_start_time and self_end_time <= interval_end_time) or (
            self_start_time >= interval_start_time and self_end_time >= interval_end_time and
            self_start_time <= interval_end_time) or (
            self_start_time <= interval_start_time and self_end_time >= interval_start_time) or (
            self_start_time <= interval_end_time and self_end_time >= interval_end_time))
        return cond

    def spans_time_point(self, point):
        return self.start_time <= point <= self.end_time


class Chromatogram(_TimeIntervalMethods):
    created_at = "new"
    glycan_composition = None

    nodes: 'ChromatogramTreeList'
    mass_shifts: List[MassShiftBase]
    used_as_mass_shift: List

    _total_intensity: float
    _neutral_mass: float
    _weighted_neutral_mass: float
    _last_neutral_mass: float
    _charge_states: Set[int]

    _start_time: float
    _end_time: float

    def __init__(self, composition, nodes=None, mass_shifts=None, used_as_mass_shift=None):
        if nodes is None:
            nodes = ChromatogramTreeList()
        if mass_shifts is None:
            mass_shifts = []
        if used_as_mass_shift is None:
            used_as_mass_shift = []

        self.nodes = nodes
        self._mass_shifts = mass_shifts
        self.used_as_mass_shift = used_as_mass_shift
        self._infer_mass_shifts()
        self._has_msms = None

        self.composition = composition
        self._total_intensity = None
        self._neutral_mass = None
        self._weighted_neutral_mass = None
        self._last_neutral_mass = 0.
        self._most_abundant_member = 0.
        self._charge_states = None
        self._retention_times = None
        self._peaks = None
        self._scan_ids = None
        self._start_time = None
        self._end_time = None

    def _invalidate(self):
        self._total_intensity = None
        self._last_neutral_mass = self._neutral_mass if self._neutral_mass is not None else 0.
        self._neutral_mass = None
        self._weighted_neutral_mass = None
        self._charge_states = None
        self._retention_times = None
        self._peaks = None
        self._scan_ids = None
        self._mass_shifts = None
        self._has_msms = None
        self._start_time = None
        self._end_time = None
        self._last_most_abundant_member = self._most_abundant_member
        self._most_abundant_member = None

    def invalidate(self):
        self._invalidate()

    def retain_most_abundant_member(self):
        self._neutral_mass = self._last_neutral_mass
        self._most_abundant_member = self._last_most_abundant_member

    @property
    def has_msms(self):
        if self._has_msms is None:
            self._has_msms = [node for node in self.nodes if node.has_msms]
        return self._has_msms

    @property
    def most_abundant_member(self):
        if self._most_abundant_member is None:
            self._most_abundant_member = max(node.max_intensity() for node in self.nodes)
        return self._most_abundant_member

    def _infer_mass_shifts(self):
        mass_shifts = set()
        for node in self.nodes:
            mass_shifts.update(node.node_types())
        if not mass_shifts:
            mass_shifts = [Unmodified]
        self._mass_shifts = list(mass_shifts)

    def __getitem__(self, i):
        return self.nodes[i]

    def __iter__(self):
        return iter(self.nodes)

    def total_signal_for(self, node_type=Unmodified):
        total = 0.
        for node in self.nodes:
            node = node._find(node_type)
            if node is not None:
                total += node._total_intensity_members()
        return total

    def mass_shift_signal_fractions(self):
        return ArithmeticMapping({
            k: self.total_signal_for(k) for k in self.mass_shifts
        })

    def drop_mass_shifts(self):
        for node in self.nodes.unspool():
            node.node_type = Unmodified
        return self

    @property
    def integrated_abundance(self):
        spacing = np.array([
            node.retention_time for node in self.nodes
        ])
        values = np.array([
            node.total_intensity() for node in self.nodes
        ])
        integrated = trapezoid(values, spacing)
        return integrated

    @property
    def mass_shifts(self):
        if self._mass_shifts is None:
            self._infer_mass_shifts()
        return self._mass_shifts

    @property
    def total_signal(self):
        if self._total_intensity is None:
            total = 0.
            for node in self.nodes:
                total += node.total_intensity()
            self._total_intensity = total
        return self._total_intensity

    @property
    def weighted_neutral_mass(self):
        if self._weighted_neutral_mass is None:
            try:
                self._infer_neutral_mass()
            except KeyError:
                self._infer_neutral_mass(self.mass_shifts[0])
        return self._weighted_neutral_mass

    @property
    def theoretical_mass(self):
        if self.composition:
            return self.composition.total_composition().mass
        else:
            return self.weighted_neutral_mass

    def _infer_neutral_mass(self, node_type=Unmodified):
        prod = 0
        total = 0
        maximum_intensity = 0
        best_neutral_mass = 0
        for node in self.nodes:
            intensity = node.max_intensity()
            try:
                mass = node.neutral_mass_of_type(node_type)
            except KeyError:
                continue
            prod += intensity * mass
            total += intensity
            if intensity > maximum_intensity:
                maximum_intensity = intensity
                best_neutral_mass = mass
        if total > 0:
            self._weighted_neutral_mass = prod / total - node_type.mass
        else:
            self._weighted_neutral_mass = best_neutral_mass - node_type.mass
        self._last_neutral_mass = self._neutral_mass = best_neutral_mass - node_type.mass
        if self._neutral_mass == 0:
            raise KeyError(node_type)
        return best_neutral_mass

    def mzs(self):
        return self._average_mz_included()

    def _average_mz_included(self):
        peaks = defaultdict(list)
        for node in self.nodes.unspool():
            for member in node.members:
                peaks[node.node_type, member.charge].append(member)
        mzs = dict()
        for key, value in peaks.items():
            product = 0
            weight = 0
            for v in value:
                product += v.mz * v.intensity
                weight += v.intensity
            mzs[key] = product / weight
        return tuple(mzs.values())

    @property
    def neutral_mass(self):
        if self._neutral_mass is None:
            try:
                self._infer_neutral_mass()
            except KeyError:
                self._infer_neutral_mass(self.mass_shifts[0])
        return self._neutral_mass

    @property
    def charge_states(self):
        if self._charge_states is None:
            states = Counter()
            for node in self.nodes:
                states += (Counter(node.charge_states()))
            # Require more than `MIN_POINTS_FOR_CHARGE_STATE` data points to accept any
            # charge state
            collapsed_states = {k for k, v in states.items() if v >= min(MIN_POINTS_FOR_CHARGE_STATE, len(self))}
            if not collapsed_states:
                collapsed_states = states.keys()
            self._charge_states = collapsed_states
        return self._charge_states

    @property
    def n_charge_states(self):
        return len(self.charge_states)

    @property
    def key(self):
        if self.composition is not None:
            return self.composition
        else:
            return self.neutral_mass

    @property
    def retention_times(self):
        if self._retention_times is None:
            self._retention_times = tuple(node.retention_time for node in self.nodes)
        return self._retention_times

    @property
    def scan_ids(self):
        if self._scan_ids is None:
            self._scan_ids = tuple(node.scan_id for node in self.nodes)
        return self._scan_ids

    @property
    def peaks(self):
        if self._peaks is None:
            self._peaks = tuple(node.peaks for node in self.nodes)
        return self._peaks

    @property
    def start_time(self):
        if self._start_time is None:
            self._start_time = self.nodes[0].retention_time
        return self._start_time

    @property
    def end_time(self):
        if self._end_time is None:
            self._end_time = self.nodes[-1].retention_time
        return self._end_time

    def as_arrays(self):
        rts = np.array([node.retention_time for node in self.nodes], dtype=np.float64)
        signal = np.array([node.total_intensity() for node in self.nodes], dtype=np.float64)
        return rts, signal

    def __len__(self):
        return len(self.nodes)

    def __repr__(self):
        return "Chromatogram(%s, %0.4f)" % (self.composition, self.weighted_neutral_mass)

    def split_sparse(self, delta_rt: float=1.) -> List['Chromatogram']:
        chunks = []
        current_chunk = []
        last_rt = self.nodes[0].retention_time

        for node in self.nodes:
            if (node.retention_time - last_rt) > delta_rt:
                x = self.__class__(self.composition, ChromatogramTreeList(current_chunk))
                x.used_as_mass_shift = list(self.used_as_mass_shift)

                chunks.append(x)
                current_chunk = []

            last_rt = node.retention_time
            current_chunk.append(node)

        x = self.__class__(self.composition, ChromatogramTreeList(current_chunk))
        x.used_as_mass_shift = list(self.used_as_mass_shift)

        chunks.append(x)
        for chunk in chunks:
            chunk.created_at = self.created_at

        # Sanity check previously done here
        # for member in chunks:
        #     for other in chunks:
        #         if member == other:
        #             continue
        #         assert not member.overlaps_in_time(other)

        return chunks

    def truncate_before(self, time: float):
        _, i = self.nodes.find_time(time)
        if self.nodes[i].retention_time < time:
            i += 1
        self.nodes = ChromatogramTreeList(self.nodes[i:])
        self._invalidate()

    def truncate_after(self, time: float):
        _, i = self.nodes.find_time(time)
        if self.nodes[i].retention_time < time:
            i += 1
        self.nodes = ChromatogramTreeList(self.nodes[:i])
        self._invalidate()

    def clone(self, cls=None):
        if cls is None:
            cls = self.__class__
        c = cls(
            self.composition, self.nodes.clone(), list(self.mass_shifts), list(self.used_as_mass_shift))
        c.created_at = self.created_at
        return c

    def insert_node(self, node: 'ChromatogramTreeNode'):
        self.nodes.insert_node(node)
        self._invalidate()

    def insert(self, scan_id: str, peak: DeconvolutedPeak, retention_time: float):
        self.nodes.insert(retention_time, scan_id, [peak])
        self._invalidate()

    def merge(self, other: 'Chromatogram', node_type=Unmodified, skip_duplicate_nodes=False):
        if skip_duplicate_nodes:
            return self._merge_missing_only(other, node_type=node_type)
        new = self.clone()
        for node in other.nodes.unspool_strip_children():
            node = node.clone()
            node.node_type = node.node_type + node_type
            new.insert_node(node)
        new.created_at = "merge"
        return new

    def deduct_node_type(self, node_type: MassShiftBase):
        new = self.clone()
        for node in new.nodes.unspool():
            if node.node_type == node_type:
                node.node_type = Unmodified
            else:
                node.node_type = node.node_type - node_type
        new.invalidate()
        return new

    def _merge_missing_only(self, other: 'Chromatogram', node_type: MassShiftBase=Unmodified):
        new = self.clone()
        ids = set(node.node_id for node in new.nodes.unspool())
        for node in other.nodes.unspool_strip_children():
            if node.node_id in ids:
                continue
            else:
                new_node = node.clone()
                new_node.node_type = node.node_type + node_type
                new.insert_node(new_node)
        new.created_at = "_merge_missing_only"
        return new

    @classmethod
    def from_parts(cls, composition, retention_times, scan_ids, peaks):
        nodes = ChromatogramTreeList()
        nodes.extend(zip(scan_ids, peaks, retention_times))
        return cls(composition, nodes)

    def slice(self, start: float, end: float):
        """
        Slice the chromatogram along the time dimension.

        Parameters
        ----------
        start : float
            The time to start from
        end : float
            The time to end at

        Returns
        -------
        Chromatogram
        """
        _, i = self.nodes.find_time(start)
        _, j = self.nodes.find_time(end)
        new = self.__class__(
            self.composition,
            ChromatogramTreeList(node.clone() for node in self.nodes[i:j + 1]),
            used_as_mass_shift=list(self.used_as_mass_shift))
        return new

    def bisect_mass_shift(self, mass_shift: MassShiftBase):
        """
        Split a chromatogram into two parts, one with the provided mass shift
        and one without.

        This method does not try to deal with composite mass shifts, see :meth:`deducte_node_type`
        for that behavior.

        Parameters
        ----------
        mass_shift : MassShift
            The mass shift to split on

        Returns
        -------
        new_with_mass_shift : Chromatogram
            The portion of the chromatogram that has the mass shift
        new_no_mass_shift : Chromatogram
            The portion of the chromatogram that did not have the mass shift

        See Also
        --------
        :meth:`deduct_node_type`
        """
        new_mass_shift = self.__class__(self.composition)
        new_no_mass_shift = self.__class__(self.composition)
        for node in self:
            for new_node in node._unspool_strip_children():
                if new_node.node_type == mass_shift:
                    new_mass_shift.insert_node(new_node)
                else:
                    new_no_mass_shift.insert_node(new_node)
        return new_mass_shift, new_no_mass_shift

    def bisect_charge(self, charge: int):
        new_charge = self.__class__(self.composition)
        new_no_charge = self.__class__(self.composition)
        for node in self.nodes.unspool():
            node_t = node.node_type
            rt = node.retention_time
            scan_id = node.scan_id
            peaks = node.members
            charge_peaks = []
            no_charge_peaks = []
            for peak in peaks:
                if peak.charge == charge:
                    charge_peaks.append(peak)
                else:
                    no_charge_peaks.append(peak)
            if len(charge_peaks):
                charge_node = ChromatogramTreeNode(
                    retention_time=rt, scan_id=scan_id, children=None,
                    members=charge_peaks, node_type=node_t)
                new_charge.insert_node(charge_node)
            if len(no_charge_peaks):
                no_charge_node = ChromatogramTreeNode(
                    retention_time=rt, scan_id=scan_id, children=None,
                    members=no_charge_peaks, node_type=node_t)
                new_no_charge.insert_node(no_charge_node)
        return new_charge, new_no_charge

    def __eq__(self, other):
        if other is None:
            return False
        if self.key != other.key:
            return False
        else:
            return self.peaks == other.peaks

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash((self.neutral_mass, self.start_time, self.end_time))

    @property
    def elemental_composition(self):
        return None

    def common_nodes(self, other: 'Chromatogram'):
        return self.nodes.common_nodes(other.nodes)

    @property
    def apex_time(self):
        rt, intensity = self.as_arrays()
        return rt[np.argmax(intensity)]

    def extract_components(self):
        mass_shifts = list(self.mass_shifts)
        labels = {}
        rest = self
        for mass_shift in mass_shifts:
            with_mass_shift, rest = rest.bisect_mass_shift(mass_shift)
            labels[mass_shift] = with_mass_shift

        labels[Unmodified] = rest
        mass_shift_charge_table = defaultdict(dict)
        for mass_shift, component in labels.items():
            charges = list(component.charge_states)
            rest = component
            for charge in charges:
                selected, rest = rest.bisect_charge(charge)
                mass_shift_charge_table[mass_shift][charge] = selected
        return mass_shift_charge_table

    def clear(self):
        self.nodes.clear()
        self._invalidate()

    def is_distinct(self, other: 'Chromatogram'):
        return not self.nodes.common_peaks(get_chromatogram(other).nodes)

    def integrate(self):
        time, intensity = self.as_arrays()
        return trapezoid(intensity, time)


class ChromatogramTreeList(object):
    roots: List['ChromatogramTreeNode']

    def __init__(self, roots=None):
        if roots is None:
            roots = []
        self.roots = list(roots)
        self._node_id_hash = None
        self._peak_hash = None

    def _invalidate(self):
        self._node_id_hash = None
        self._peak_hash = None

    def find_time(self, retention_time: float):
        if len(self.roots) == 0:
            raise EmptyListException()
        lo = 0
        hi = len(self.roots)
        while lo != hi:
            i = (lo + hi) // 2
            node = self.roots[i]
            if node.retention_time == retention_time:
                return node, i
            elif (hi - lo) == 1:
                return None, i
            elif node.retention_time < retention_time:
                lo = i
            elif node.retention_time > retention_time:
                hi = i

    def _build_node_id_hash(self):
        node_id_hash = set()
        for node in self.unspool():
            node_id_hash.add(node.node_id)
        self._node_id_hash = frozenset(node_id_hash)

    def _build_peak_hash(self):
        peak_hash = set()
        for node in self.unspool():
            peak_hash.update([(node.scan_id, peak) for peak in node.members])
        self._peak_hash = frozenset(peak_hash)

    @property
    def node_id_hash(self):
        if self._node_id_hash is None:
            self._build_node_id_hash()
        return self._node_id_hash

    @property
    def peak_hash(self):
        if self._peak_hash is None:
            self._build_peak_hash()
        return self._peak_hash

    def insert_node(self, node: 'ChromatogramTreeNode'):
        self._invalidate()
        try:
            root, i = self.find_time(node.retention_time)
            if root is None:
                if i != 0:
                    self.roots.insert(i + 1, node)
                else:
                    slot = self.roots[i]
                    if slot.retention_time < node.retention_time:
                        i += 1
                    self.roots.insert(i, node)
            else:
                root.add(node)
            return i
        except EmptyListException:
            self.roots.append(node)
            return 0

    def insert(self, retention_time, scan_id, peaks, node_type=Unmodified):
        node = ChromatogramTreeNode(retention_time, scan_id, [], peaks, node_type)
        return self.insert_node(node)

    def extend(self, iterable: Iterable['ChromatogramTreeNode']):
        for scan_id, peaks, retention_time in iterable:
            self.insert(retention_time, scan_id, peaks)

    def __getitem__(self, i) -> 'ChromatogramTreeNode':
        return self.roots[i]

    def __len__(self):
        return len(self.roots)

    def __iter__(self):
        return iter(self.roots)

    def __eq__(self, other):
        return list(self) == list(other)

    def __ne__(self, other):
        return not (self == other)

    def clone(self):
        return ChromatogramTreeList(node.clone() for node in self)

    def clear(self):
        self.roots = []
        self._invalidate()

    def unspool(self):
        out_queue = []
        for root in self:
            stack = [root]
            while len(stack) != 0:
                node = stack.pop()
                out_queue.append(node)
                stack.extend(node.children)
        return out_queue

    def unspool_strip_children(self) -> List['ChromatogramTreeNode']:
        out_queue = []
        for root in self:
            stack = [root]
            while len(stack) != 0:
                node = stack.pop()
                node_copy = node.clone()
                node_copy.children = []
                out_queue.append(node_copy)
                stack.extend(node.children)
        return out_queue

    def common_nodes(self, other):
        return not self.node_id_hash.isdisjoint(other.node_id_hash)

    def common_peaks(self, other):
        return not self.peak_hash.isdisjoint(other.peak_hash)

    def __repr__(self):
        return "ChromatogramTreeList(%d nodes, %0.2f-%0.2f)" % (
            len(self), self[0].retention_time, self[-1].retention_time)


class ChromatogramTreeNode(object):
    __slots__ = ['retention_time', 'scan_id', 'children', 'members', 'node_type',
                 '_most_abundant_member', '_neutral_mass', '_charge_states', '_has_msms',
                 'node_id', ]

    retention_time: float
    scan_id: str
    children: List['ChromatogramTreeNode']
    members: List[DeconvolutedPeak]
    node_type: MassShiftBase
    node_id: int

    _most_abundant_member: Optional[DeconvolutedPeak]
    _neutral_mass: float
    _charge_states: Set[int]
    _has_msms: Optional[bool]

    def __init__(self, retention_time=None, scan_id=None, children=None, members=None,
                 node_type=Unmodified):
        if children is None:
            children = []
        if members is None:
            members = []
        self.retention_time = retention_time
        self.scan_id = scan_id
        self.children = children
        self.members = members
        self.node_type = node_type
        self._most_abundant_member = None
        self._neutral_mass = 0
        self._charge_states = set()
        self._recalculate()
        self._has_msms = None
        self.node_id = uid()

    def as_ref(self):
        return {
            "scan_id": self.scan_id,
            "node_id": self.node_id,
            "peak_indices": [p.index.neutral_mass for p in self.members],
            "node_type": self.node_type.name,
            "children": [
                node._as_ref() for node in self.children
            ]
        }

    @classmethod
    def from_ref(cls, ref: dict, scan_reader: ProcessedRandomAccessScanSource):
        scan = scan_reader.get_scan_by_id(ref['scan_id'])
        rt = scan.scan_time
        members = [scan.deconvoluted_peak_set[i] for i in ref['peak_indices']]
        node_type = MassShiftBase.get(ref['node_type'])
        children = [cls.from_ref(child) for child in ref['children']]
        node = cls(rt, ref['scan_id'], children, members, node_type)
        node.node_id = ref['node_id']
        return node

    def clone(self):
        node = ChromatogramTreeNode(
            self.retention_time, self.scan_id, [c.clone() for c in self.children],
            list(self.members), node_type=self.node_type)
        node.node_id = self.node_id
        return node

    def __reduce__(self):
        return self.__class__, (
            self.retention_time, self.scan_id, [c for c in self.children],
            list(self.members), self.node_type), self.__getstate__()

    def __getstate__(self):
        return {
            "node_id": self.node_id
        }

    def __setstate__(self, state):
        self.node_id = state['node_id']

    def _unspool_strip_children(self):
        node = ChromatogramTreeNode(
            self.retention_time, self.scan_id, [], list(self.members), node_type=self.node_type)
        yield node
        for child in self.children:
            for node in child._unspool_strip_children():
                yield node

    def _calculate_most_abundant_member(self):
        if len(self.members) == 1:
            self._most_abundant_member = self.members[0]
        else:
            if len(self.members) == 0:
                self._most_abundant_member = None
            else:
                self._most_abundant_member = max(self.members, key=intensity_getter)

    def _recalculate(self):
        self._calculate_most_abundant_member()
        self._neutral_mass = self._most_abundant_member.neutral_mass
        self._charge_states = None
        self._has_msms = None

    @property
    def _contained_charge_states(self):
        if self._charge_states is None:
            self._charge_states = set(split_by_charge(self.members))
        return self._charge_states

    @property
    def has_msms(self):
        if self._has_msms is None:
            self._has_msms = self._has_any_peaks_with_msms()
        return self._has_msms

    def _find(self, node_type=Unmodified):
        if self.node_type == node_type:
            return self
        else:
            for child in self.children:
                match = child._find(node_type)
                if match is not None:
                    return match

    def find(self, node_type=Unmodified):
        match = self._find(node_type)
        if match is not None:
            return match
        else:
            raise KeyError(node_type)

    @property
    def neutral_mass(self):
        if self._neutral_mass is None:
            if self._most_abundant_member is not None:
                self._neutral_mass = self._most_abundant_member.neutral_mass
        return self._neutral_mass

    def charge_states(self):
        u = set()
        u.update(self._contained_charge_states)
        for child in self.children:
            u.update(child.charge_states())
        return u

    def neutral_mass_of_type(self, node_type=Unmodified):
        return self.find(node_type)._neutral_mass

    def add(self, node, recalculate=True):
        if self.node_id == node.node_id:
            raise DuplicateNodeError("Duplicate Node %s" % node)
        if node.node_type == self.node_type:
            self.members.extend(node.members)
        else:
            self.children.append(node)
        if recalculate:
            self._recalculate()

    def _total_intensity_members(self):
        total = 0.
        for peak in self.members:
            total += peak.intensity
        return total

    def _total_intensity_children(self):
        total = 0.
        for child in self.children:
            total += child.total_intensity()
        return total

    def max_intensity(self):
        return self._most_abundant_member.intensity

    def total_intensity(self):
        return self._total_intensity_children() + self._total_intensity_members()

    def __eq__(self, other):
        if self.scan_id != other.scan_id:
            return False
        return self.members == other.members

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.node_id)

    def _has_any_peaks_with_msms(self):
        for peak in self.members:
            if peak.chosen_for_msms:
                return True
        for child in self.children:
            if child._has_any_peaks_with_msms():
                return True
        return False

    @property
    def peaks(self):
        peaks = list(self.members)
        for child in self.children:
            peaks.extend(child.peaks)
        return peaks

    def __repr__(self):
        return "ChromatogramTreeNode(%f, %r, %s|%d, %d)" % (
            self.retention_time, self.scan_id, self.node_type.name,
            len(self.members), len(self.children))

    def node_types(self):
        kinds = [self.node_type]
        for child in self.children:
            kinds.extend(child.node_types())
        return kinds


@add_metaclass(ABCMeta)
class ChromatogramInterface(object):
    pass


ChromatogramInterface.register(Chromatogram)


class ChromatogramWrapper(_TimeIntervalMethods):
    def __init__(self, chromatogram):
        self.chromatogram = chromatogram

    def __iter__(self):
        return iter(self.chromatogram)

    def __getitem__(self, i):
        return self.chromatogram[i]

    def __len__(self):
        return len(self.chromatogram)

    def __hash__(self):
        return hash(self.chromatogram)

    def __eq__(self, other):
        try:
            return self.chromatogram == get_chromatogram(other)
        except Exception:
            return False

    def has_chromatogram(self):
        return self.chromatogram is not None

    @property
    def nodes(self):
        return self.chromatogram.nodes

    @property
    def neutral_mass(self):
        return self.chromatogram.neutral_mass

    @property
    def weighted_neutral_mass(self):
        return self.chromatogram.weighted_neutral_mass

    @property
    def theoretical_mass(self):
        return self.chromatogram.theoretical_mass

    @property
    def n_charge_states(self):
        return self.chromatogram.n_charge_states

    @property
    def charge_states(self):
        return self.chromatogram.charge_states

    @property
    def mass_shifts(self):
        return self.chromatogram.mass_shifts

    @property
    def total_signal(self):
        return self.chromatogram.total_signal

    @property
    def integrated_abundance(self):
        return self.chromatogram.integrated_abundance

    @property
    def start_time(self):
        return self.chromatogram.start_time

    @property
    def end_time(self):
        return self.chromatogram.end_time

    def as_arrays(self):
        return self.chromatogram.as_arrays()

    def overlaps_in_time(self, interval):
        return self.chromatogram.overlaps_in_time(interval)

    @property
    def key(self):
        return self.chromatogram.key

    @property
    def composition(self):
        return self.chromatogram.composition

    @property
    def peaks(self):
        return self.chromatogram.peaks

    @property
    def scan_ids(self):
        return self.chromatogram.scan_ids

    @property
    def retention_times(self):
        return self.chromatogram.retention_times

    def __repr__(self):
        return "{self.__class__.__name__}({self.key}, {self.neutral_mass})".format(self=self)

    @property
    def entity(self):
        return self.chromatogram.entity

    @entity.setter
    def entity(self, value):
        self.chromatogram.entity = value

    @property
    def glycan_composition(self):
        return self.chromatogram.glycan_composition

    @property
    def elemental_composition(self):
        return self.chromatogram.elemental_composition

    def common_nodes(self, other):
        return self.chromatogram.common_nodes(other)

    @property
    def apex_time(self):
        return self.chromatogram.apex_time

    def __getattr__(self, name):
        if name == 'chromatogram':
            raise AttributeError(name)
        else:
            return getattr(self.chromatogram, name)

    def clone(self):
        chromatogram = self.chromatogram.clone()
        new = self.__class__(chromatogram)
        return new

    def clear(self):
        self.chromatogram.clear()

    def mzs(self):
        return self.chromatogram.mzs()

    def is_distinct(self, other):
        return self.chromatogram.is_distinct(get_chromatogram(other))

    def mass_shift_signal_fractions(self):
        return self.chromatogram.mass_shift_signal_fractions()

    def drop_mass_shifts(self):
        self.chromatogram.drop_mass_shifts()
        return self

    def integrate(self):
        return self.chromatogram.integrate()


ChromatogramInterface.register(ChromatogramWrapper)


class CachedGlycanComposition(HashableGlycanComposition):
    _hash = None

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(str(self))
        return self._hash


class EntityChromatogram(Chromatogram):
    _entity = None

    @property
    def glycan_composition(self):
        return self._entity

    def clone(self, cls=None):
        if cls is None:
            cls = self.__class__
        inst = super(EntityChromatogram, self).clone(cls=cls)
        inst.entity = self.entity
        return inst

    @property
    def entity(self):
        if self._entity is None and self.composition is not None:
            self.entity = self.composition
        return self._entity

    @entity.setter
    def entity(self, value):
        if isinstance(value, str):
            value = self._parse(value)
        self._entity = value
        if self.composition is None and value is not None:
            self.composition = value

    @property
    def structure(self):
        return self.entity

    @structure.setter
    def structure(self, value):
        self.entity = value

    @property
    def elemental_composition(self):
        try:
            return self.entity.total_composition()
        except AttributeError:
            try:
                return self._parse(self.composition).total_composition()
            except Exception:
                return None


class GlycanCompositionChromatogram(EntityChromatogram):
    @staticmethod
    def _parse(string):
        return CachedGlycanComposition.parse(string)

    @property
    def glycan_composition(self):
        if isinstance(self.composition, str):
            self.composition = self._parse(self.composition)
        return self.composition


class GlycopeptideChromatogram(EntityChromatogram):
    @staticmethod
    def _parse(string):
        from glycopeptidepy.structure.sequence import PeptideSequence
        return PeptideSequence(string)

    @property
    def glycan_composition(self):
        if self.entity is None and self.composition is not None:
            self.entity = self.composition
        return self.entity.glycan_composition


def get_chromatogram(instance):
    if isinstance(instance, Chromatogram):
        return instance
    elif isinstance(instance, ChromatogramInterface):
        return instance
    elif isinstance(instance, ChromatogramWrapper):
        return instance.chromatogram
    else:
        try:
            return instance.get_chromatogram()
        except AttributeError:
            raise TypeError(
                "%s does not contain a chromatogram" % instance)



from glypy.utils import Enum


class SpectrumMatchClassification(Enum):
    target_peptide_target_glycan = 0
    target_peptide_decoy_glycan = 1
    decoy_peptide_target_glycan = 2
    decoy_peptide_decoy_glycan = decoy_peptide_target_glycan | target_peptide_decoy_glycan


class AnalysisTypeEnum(Enum):
    glycopeptide_lc_msms = 0
    glycan_lc_ms = 1


class GlycopeptideSearchStrategy(Enum):
    target_internal_decoy_competition = "target-internal-decoy-competition"
    target_decoy_competition = "target-decoy-competition"
    multipart_target_decoy_competition = "multipart-target-decoy-competition"


AnalysisTypeEnum.glycopeptide_lc_msms.add_name("Glycopeptide LC-MS/MS")
AnalysisTypeEnum.glycan_lc_ms.add_name("Glycan LC-MS")


class GlycopeptideFDREstimationStrategy(Enum):
    multipart_gamma_gaussian_mixture = 0
    peptide_fdr = 1
    glycan_fdr = 2
    glycopeptide_fdr = 3
    peptide_or_glycan = 4


GlycopeptideFDREstimationStrategy.multipart_gamma_gaussian_mixture.add_name(
    "multipart")
GlycopeptideFDREstimationStrategy.multipart_gamma_gaussian_mixture.add_name(
    "joint")
GlycopeptideFDREstimationStrategy.peptide_fdr.add_name("peptide")
GlycopeptideFDREstimationStrategy.glycan_fdr.add_name("glycan")
GlycopeptideFDREstimationStrategy.peptide_or_glycan.add_name('any')



from .hypothesis import GlycanHypothesis, GlycopeptideHypothesis

from .glycan import (
    GlycanComposition, GlycanCombination, GlycanClass,
    GlycanStructure, GlycanTypes, GlycanCombinationGlycanComposition,
    GlycanCompositionToClass, GlycanStructureToClass)

from .peptide import (
    Protein, Peptide, Glycopeptide, ProteinSite)

from .generic import (
    TemplateNumberStore as TemplateNumberStore,
    ReferenceDatabase, ReferenceAccessionNumber, FileBlob)



import math

from array import array
from typing import Any, DefaultDict, List, Tuple, Callable, NamedTuple, Union, Dict, Optional, Counter
from glycresoft.structure.scan import ScanStub, ScanWrapperBase, ScanInformation
from glycresoft.tandem.ref import SpectrumReference

from ms_deisotope.data_source import ProcessedScan
from ms_deisotope.output import ProcessedMzMLLoader

from glycopeptidepy.structure import PeptideSequence, Modification, ModificationRule

from glycresoft.task import TaskBase
from glycresoft.structure import FragmentCachingGlycopeptide
from glycresoft.structure.probability import KDModel, MultiKDModel

from glycresoft.tandem.peptide.scoring import localize

from glycresoft.tandem.spectrum_match import SpectrumMatch, SpectrumSolutionSet, LocalizationScore

from glypy.structure.glycan_composition import HashableGlycanComposition


class PeptideGroupToken(NamedTuple):
    peptide: PeptideSequence
    modifications: Tuple[Tuple[Union[Modification, ModificationRule], int]]
    glycan_composition: HashableGlycanComposition

    def __eq__(self, other: 'PeptideGroupToken'):
        if not self.peptide.base_sequence_equality(other.peptide):
            return False
        if self.glycan_composition != other.glycan_composition:
            return False
        if self.modifications != other.modifications:
            return False
        return True

    def __ne__(self, other: 'PeptideGroupToken'):
        return not self == other

    def __hash__(self):
        code = hash(self.glycan_composition) & hash(self.modifications)
        code &= len(self.peptide)
        code &= hash(self.peptide[0].symbol)
        return code


SolutionsBin = Tuple[List[SpectrumMatch], PeptideGroupToken]
EvaluatedSolutionBins = Tuple[List[SpectrumMatch],
                              List[localize.PTMProphetEvaluator]]

ScanOrRef = Union[ProcessedScan, ScanStub,
                  SpectrumReference, ScanWrapperBase, ScanInformation]


class LocalizationGroup(NamedTuple):
    spectrum_matches: List[SpectrumMatch]
    localization_matches: List[localize.PTMProphetEvaluator]


class EvaluatedSolutionBins(NamedTuple):
    solution_set: SpectrumSolutionSet
    groups: List[LocalizationGroup]

    @property
    def scan_id(self):
        return self.solution_set.scan.scan_id


class ModificationLocalizationSearcher(TaskBase):
    error_tolerance: float
    threshold_fn: Callable[[SpectrumMatch], bool]
    restricted_modifications: Dict[str, ModificationRule]
    model: Optional[MultiKDModel]

    def __init__(self,
                 threshold_fn: Callable[[SpectrumMatch],
                                        bool] = lambda x: x.q_value < 0.05,
                 error_tolerance: float = 2e-5,
                 restricted_modifications: Optional[Dict[str, ModificationRule]] = None,
                 model: Optional[MultiKDModel]=None):
        if restricted_modifications is None:
            restricted_modifications = {}
        self.threshold_fn = threshold_fn
        self.error_tolerance = error_tolerance
        self.restricted_modifications = restricted_modifications
        self.model = model

    def get_modifications_for_peptide(self, peptide: FragmentCachingGlycopeptide) -> Tuple[Tuple[Tuple[Modification, int]],
                                                                                           HashableGlycanComposition]:
        modifications = Counter()
        for position in peptide:
            if position.modifications:
                mod = position.modifications[0].rule
                if mod.name in self.restricted_modifications:
                    mod = self.restricted_modifications[mod.name]
                modifications[mod] += 1
        modifications = list(modifications.items())
        modifications.sort(key=lambda x: x[0].name)
        glycan = None
        if peptide.glycan_composition:
            glycan = peptide.glycan_composition
        return tuple(modifications), glycan

    def find_overlapping_localization_solutions(self, solution_set: SpectrumSolutionSet) -> List[Tuple[List[SpectrumMatch],
                                                                                                       PeptideGroupToken]]:
        if isinstance(solution_set, SpectrumMatch):
            solution_set = [solution_set]
        bins: List[Tuple[List[SpectrumMatch],
                         PeptideGroupToken]] = []
        for sm in solution_set:
            if not self.threshold_fn(sm):
                continue
            target: FragmentCachingGlycopeptide = sm.target
            mod_signature, glycan = self.get_modifications_for_peptide(target)
            token = PeptideGroupToken(target, mod_signature, glycan)
            for solution_bin, bin_token in bins:
                if bin_token == token:
                    solution_bin.append(sm)
                    break
            else:
                bins.append(([sm], token))
        return bins

    def process_localization_bin(self,
                                 scan: ProcessedScan,
                                 modification_group_token: PeptideGroupToken) -> List[localize.PTMProphetEvaluator]:
        solutions = []
        scan_description = ScanInformation.from_scan(scan)
        for mod_sig, count in modification_group_token.modifications:
            el = localize.PTMProphetEvaluator(
                scan,
                modification_group_token.peptide,
                modification_rule=mod_sig,
                modification_count=count
            )
            el.score_arrangements(
                error_tolerance=self.error_tolerance
            )
            # At this point we don't need the full scan representation anymore,
            # so we can drop the peaks later.
            el.scan = scan_description
            solutions.append(el)
        return solutions

    def get_training_instances(self, solutions: List[EvaluatedSolutionBins]) -> List[localize.PTMProphetEvaluator]:
        candidates = []
        for sset_bin in solutions:
            for sol in sset_bin.groups:
                for loc in sol.localization_matches:
                    if len(loc.peptidoforms) > 1:
                        candidates.append(loc)
        return candidates

    def train_ptm_prophet(self, training_instances: List[localize.PTMProphetEvaluator],
                          maxiter: int = 150) -> MultiKDModel:
        o_scores = array('d')
        m_scores = array('d')
        for inst in training_instances:
            for scores in inst.solution_for_site.values():
                if not math.isfinite(scores.o_score) or not math.isfinite(scores.m_score):
                    continue
                o_scores.append(scores.o_score)
                m_scores.append(scores.m_score)

        while len(o_scores) < 2:
            o_scores.append(0.99)
            m_scores.append(0.99)
        o_scores.append(0.5)
        m_scores.append(0.5)

        prophet = MultiKDModel(0.5, [KDModel(), KDModel()])
        for o, m in zip(o_scores, m_scores):
            prophet.add(o, [o, m])

        self.log("Begin fitting localization score model")
        try:
            it, delta = prophet.fit(maxiter)
        except ValueError as err:
            self.log(f"Failed to fit localization model: {err}")
            self.model = None
            return None
        if it < maxiter:
            self.log(
                f"Localization score model converged after {it} iterations")
        else:
            self.log(
                f"The localization model failed to converge after {it} iterations ({delta})")
        self.model = prophet
        if prophet is not None:
            prophet.close_thread_pool()
        return prophet

    def _select_top_isoform_in_bin(self, sol: LocalizationGroup, prophet: Optional[MultiKDModel]=None):
        isoform_scores: DefaultDict[str,
                                    List[LocalizationScore]] = DefaultDict(list)
        isoform_weights: DefaultDict[str, float] = DefaultDict(float)
        for loc in sol.localization_matches:
            seen = set()
            for iso, scores, weight in loc.score_isoforms(prophet=prophet):
                key = str(iso.peptide)
                if key in seen:
                    continue
                isoform_scores[key].extend(scores)
                isoform_weights[key] += weight

        max_weight = max(isoform_weights.values())
        for psm in sol.spectrum_matches:
            key = str(psm.target)
            if key in isoform_scores:
                psm.localizations = isoform_scores[key]
            if abs(isoform_weights[key] - max_weight) > 1e-2 and psm.best_match:
                psm.best_match = False
                psm.valid = False
                self.log(f"... Invalidating {psm.target}@{psm.scan_id}")

    def select_top_isoforms(self, solutions: List[List[EvaluatedSolutionBins]], prophet: Optional[MultiKDModel]=None):
        if prophet is None:
            prophet = self.model
        for sset_bin in solutions:
            for sol in sset_bin.groups:
                self._select_top_isoform_in_bin(sol, prophet=prophet)

    def process_solution_set(self, solution_set: SpectrumSolutionSet) -> EvaluatedSolutionBins:
        solution_bins = self.find_overlapping_localization_solutions(
            solution_set)
        scan = self.resolve_spectrum(solution_set)
        solutions = [LocalizationGroup(solution_bin, self.process_localization_bin(scan, signature))
                     for solution_bin, signature in solution_bins]
        return EvaluatedSolutionBins(solution_set, solutions)

    def resolve_spectrum(self, scan_ref: ScanOrRef) -> ProcessedScan:
        if isinstance(scan_ref, ScanWrapperBase):
            scan_ref = scan_ref.scan
        if isinstance(scan_ref, ProcessedScan):
            return scan_ref
        raise TypeError(type(scan_ref))

    def with_scan_loader(self, scan_loader):
        dup = ScanLoadingModificationLocalizationSearcher(
            scan_loader, self.threshold_fn, self.error_tolerance, self.restricted_modifications)
        dup.model = self.model
        return dup



class ScanLoadingModificationLocalizationSearcher(ModificationLocalizationSearcher):
    scan_loader: ProcessedMzMLLoader

    def __init__(self, scan_loader: ProcessedMzMLLoader,
                 threshold_fn: Callable[[SpectrumMatch], bool]=lambda x: x.q_value < 0.05,
                 error_tolerance: float = 2e-5,
                 restricted_modifications: Optional[Dict[str, ModificationRule]] = None,
                 model: Optional[MultiKDModel]=None):
        self.scan_loader = scan_loader
        super().__init__(threshold_fn, error_tolerance, restricted_modifications, model)

    def __reduce__(self):
        return self.__class__, (None, self.threshold_fn, self.error_tolerance, self.restricted_modifications)

    def resolve_spectrum(self, scan_ref: ScanOrRef) -> ProcessedScan:
        if isinstance(scan_ref, ScanWrapperBase):
            scan_ref = scan_ref.scan
        if isinstance(scan_ref, ProcessedScan):
            return scan_ref
        if self.scan_loader is None:
            raise TypeError("Cannot load spectrum by reference when the `scan_loader` attribute is not set")
        return self.scan_loader.get_scan_by_id(scan_ref.scan_id)

    def simplify(self):
        return ModificationLocalizationSearcher(
            self.threshold_fn,
            self.error_tolerance,
            self.restricted_modifications,
            self.model)



from collections import defaultdict

import numpy as np
import matplotlib
from matplotlib import pyplot as plt


def _make_color_map():
    # Derived from the seaborn.cubehelix_palette function
    start = .5
    rot = - .75
    gamma = 1.0
    hue = 0.8
    light = 0.85
    dark = 0.15
    cdict = matplotlib._cm.cubehelix(gamma, start, rot, hue)
    cmap = matplotlib.colors.LinearSegmentedColormap("cubehelix", cdict)
    x_256 = np.linspace(light, dark, 256)
    pal_256 = cmap(x_256)
    cmap = matplotlib.colors.ListedColormap(pal_256)
    return cmap


_color_map = _make_color_map()


def extract_intensity_array(peaks):
    mzs = []
    intensities = []
    rts = []
    current_mzs = []
    current_intensities = []
    last_time = None
    for peak, time in peaks:
        if time != last_time:
            if last_time is not None:
                mzs.append(current_mzs)
                intensities.append(current_intensities)
                rts.append(time)
            last_time = time
            current_mzs = []
            current_intensities = []
        current_mzs.append(peak.mz)
        current_intensities.append(peak.intensity)
    mzs.append(current_mzs)
    intensities.append(current_intensities)
    rts.append(time)
    return mzs, intensities, rts


def binner(x):
    return np.floor(np.array(x) / 10.) * 10


def make_map(mzs, intensities):
    binned_mzs = [
        binner(mz_row) for mz_row in mzs
    ]
    unique_mzs = set()
    map(unique_mzs.update, binned_mzs)
    unique_mzs = np.array(sorted(unique_mzs))
    assigned_bins = []
    j = 0
    for mz, inten in zip(mzs, intensities):
        j += 1
        mz = binner(mz)
        bin_row = defaultdict(float)
        for i in range(len(mz)):
            k = mz[i]
            v = inten[i]
            bin_row[k] += v
        array_row = np.array([bin_row[m] for m in unique_mzs])
        assigned_bins.append(array_row)

    assigned_bins = np.vstack(assigned_bins)
    assigned_bins[assigned_bins == 0] = 1
    return assigned_bins, unique_mzs


def render_map(assigned_bins, rts, unique_mzs, ax=None, color_map=None, scaler=np.sqrt):
    if ax is None:
        fig, ax = plt.subplots(1)
    if color_map is None:
        color_map = _color_map
    ax.pcolormesh(np.array(scaler(assigned_bins.T)), cmap=color_map)

    xticks = ax.get_xticks()
    newticks = xticks
    newlabels = np.array(ax.get_xticklabels())[np.arange(0, len(xticks))]
    n = len(newlabels)
    step = len(rts) / n
    interp = [rts[i * step] for i in range(n)]

    for i, label in enumerate(newlabels):
        num = round(interp[i], 1)
        label.set_rotation(90)
        label.set_text(str((num)))

    ax.set_xticks(newticks)
    ax.set_xticklabels(newlabels)

    yticks = ax.get_yticks()
    n = len(yticks)

    newticks = yticks[np.arange(0, len(yticks))]
    newlabels = np.array(ax.get_yticklabels())[np.arange(0, len(yticks))]
    va = unique_mzs
    n = len(newlabels)
    step = len(va) / n
    interp = [va[i * step] for i in range(n)]
    for i, label in enumerate(newlabels):
        num = interp[i]
        label.set_text(str((num)))
    ax.set_yticks(newticks)
    ax.set_yticklabels(newlabels)

    ax.set_xlabel("Retention Time")
    ax.set_ylabel("m/z")
    return ax


def get_peak_time_pairs(peak_loader, threshold=None):
    if threshold is None:
        acc = []
        for scan_id in peak_loader.extended_index.ms1_ids:
            header = peak_loader.get_scan_header_by_id(scan_id)
            acc.extend(header.arrays[1])
        threshold = np.percentile(acc, 90)
    peaks = []
    for scan_id in peak_loader.extended_index.ms1_ids:
        scan = peak_loader.get_scan_by_id(scan_id)
        for peak in scan.deconvoluted_peak_set:
            if peak.intensity > threshold:
                peaks.append((peak, scan.scan_time))
    return peaks


class LCMSMapArtist(object):

    def __init__(self, peak_time_pairs, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1)
        self.peak_time_pairs = peak_time_pairs
        self.ax = ax

    def format_axes(self, axis_font_size=12):
        self.ax.axes.spines['right'].set_visible(False)
        self.ax.axes.spines['top'].set_visible(False)
        self.ax.yaxis.tick_left()
        self.ax.xaxis.tick_bottom()
        [t.set(fontsize=axis_font_size) for t in self.ax.get_xticklabels()]
        [t.set(fontsize=axis_font_size) for t in self.ax.get_yticklabels()]

    def draw(self):
        mzs, intensities, rts = extract_intensity_array(self.peak_time_pairs)
        binned_intensities, unique_mzs = make_map(mzs, intensities)
        render_map(binned_intensities, rts, unique_mzs, self.ax)
        self.format_axes()
        return self

    @classmethod
    def from_peak_loader(cls, peak_loader, threshold=None, ax=None):
        return cls(get_peak_time_pairs(peak_loader, threshold=threshold), ax=ax)



cimport cython
# from cpython.list cimport PyList_Append, PyList_Size, PyList_GetItem

import numpy as np
cimport numpy as np

from ms_deisotope._c.peak_set cimport DeconvolutedPeak
from glypy.composition.ccomposition cimport CComposition
from glycresoft._c.chromatogram_tree.mass_shift cimport MassShiftBase, MassShift


cdef MassShiftBase Unmodified


cdef class ChromatogramTreeNode(object):

    cdef:
        public list children
        public list members
        public double retention_time
        public object scan_id
        public MassShiftBase node_type
        public DeconvolutedPeak _most_abundant_member
        public double _neutral_mass
        public set _charge_states
        public object _has_msms
        public object node_id

    cpdef ChromatogramTreeNode clone(self)
    cpdef _recalculate(self)

    cdef set get_charge_states(self)
    cdef set get_contained_charge_states(self)

    cpdef set charge_states(self)

    cdef double get_neutral_mass(self)

    cpdef _calculate_most_abundant_member(self)

    cpdef ChromatogramTreeNode _find(self, MassShiftBase node_type=*)
    cpdef ChromatogramTreeNode find(self, MassShiftBase node_type=*)

    cdef double _total_intensity_members(self)
    cdef double _total_intensity_children(self)

    cpdef double max_intensity(self)
    cpdef double total_intensity(self)




'''Types for managing the evaluation of spectrum matches, to
describe matches in bulk for serialization.
'''
import time
from collections import defaultdict

from typing import List, Mapping, Tuple, Any, Type

from ms_deisotope.data_source import ProcessedScan

from glycresoft.task import TaskBase

from glycresoft.chromatogram_tree import MassShift

from ..spectrum_match import SpectrumMatch, MultiScoreSpectrumMatch, ScoreSet
from .task import TaskDeque


class SpectrumEvaluatorBase(object):
    """Abstract base class for abstracting away how spectrum matches are
    set up.

    Attributes
    ----------
    scan_map: Mapping
        A mapping from scan id to :class:`~.ProcessedScan`
    mass_shift_map: Mapping
        A mapping from mass shift name to :class:`~.MassShift`
    solution_map: Mapping
        A mapping from (scan id, mass shift name) to the packaged match.
    """
    scan_map: Mapping[str, ProcessedScan]
    mass_shift_map: Mapping[str, MassShift]
    solution_map: Mapping[Tuple[str, str], Any]

    def fetch_scan(self, key):
        return self.scan_map[key]

    def fetch_mass_shift(self, key):
        return self.mass_shift_map[key]

    def create_evaluation_context(self, group):
        return None

    def evaluate(self, scan: ProcessedScan, structure,
                 evaluation_context=None, *args, **kwargs):
        """Evaluate the match between ``structure`` and ``scan``

        Parameters
        ----------
        scan : ms_deisotope.ProcessedScan
            MSn scan to match against
        structure : object
            Structure to match against ``scan``
        evaluation_context: object, optional
            The evaluation context used to carry external information into the evaluation.
        *args
            Propagated to scoring function
        **kwargs
            Propagated to scoring function

        Returns
        -------
        SpectrumMatcherBase
        """
        raise NotImplementedError()

    def handle_instance(self, structure, scan, mass_shift, evaluation_context=None):
        solution = self.evaluate(scan, structure, mass_shift=mass_shift,
                                 evaluation_context=evaluation_context,
                                 **self.evaluation_args)
        self.solution_map[scan.id, mass_shift.name] = self.solution_packer(solution)
        return solution

    def construct_cache_subgroups(self, work_order):
        '''Build groups of structures which should be evaluated within the same context.
        The default implementation assumes each structure should get its own context.

        Subclasses should override this to determine how to properly group things to
        share context.

        Parameters
        ----------
        work_order: :class:`dict`
            A work order specification. Assumes the first entry in each order is the
            structure to be grouped.

        Returns
        -------
        subgroups: :class:`list` of :class:`list`
            Each distinct subgroup to be evauluated together.
        '''
        subgroups = []
        for key, order in work_order['work_orders'].items():
            record = order[0]
            subgroups.append([record])
        return subgroups

    def evaluate_subgroup(self, work_order, subgroup):
        """Evaluate a sub-group of structures with shared context against their respective
        spectra.

        Parameters
        ----------
        work_order : :class:`dict`
            The work order specification which includes the mapping from structure ID to
            spectrum hit type pairs.
        subgroup : :class:`list`
            The list of structures to evaluate together

        Returns
        -------
        results: :class:`list`
            A list of packed results for each member of the group
        """
        results = []
        evaluation_context = self.create_evaluation_context(subgroup)
        for structure in subgroup:
            scan_specification = work_order['work_orders'][structure.id][1]
            scan_specification = [
                (self.fetch_scan(i), self.fetch_mass_shift(j)) for i, j in scan_specification]
            solution_target = None
            solution = None
            for scan, mass_shift in scan_specification:
                solution = self.handle_instance(
                    structure, scan, mass_shift, evaluation_context)
                solution_target = solution.target
            if solution is not None:
                try:
                    solution.target.clear_caches()
                except AttributeError:
                    pass
            packed = self.pack_output(solution_target)
            results.append(packed)
        return results

    def handle_group(self, work_order):
        results = []
        hit_cache_groups = self.construct_cache_subgroups(work_order)
        for subgroup in hit_cache_groups:
            results.extend(self.evaluate_subgroup(work_order, subgroup))
        return results

    def handle_item(self, structure, scan_specification):
        scan_specification = [(self.fetch_scan(i), self.fetch_mass_shift(j)) for i, j in scan_specification]
        solution_target = None
        solution = None
        for scan, mass_shift in scan_specification:
            solution = self.handle_instance(structure, scan, mass_shift)
            solution_target = solution.target

        if solution is not None:
            try:
                solution.target.clear_caches()
            except AttributeError:
                pass
        return self.pack_output(solution_target)

    def pack_output(self, target):
        raise NotImplementedError()


class LocalSpectrumEvaluator(SpectrumEvaluatorBase, TaskBase):
    '''A :class:`SpectrumEvaluatorBase` implementation that counter-points
    :class:`IdentificationProcessDispatcher` by doing all of its evaluation within
    a single process.

    To avoid excessive subclassing, instances first try to steal the
    :meth:`construct_cache_subgroups` and :meth:`create_evaluation_context` from
    :attr:`evaluator`.
    '''

    log_process_cycle = 5000

    evaluator: SpectrumEvaluatorBase
    evaluation_args: dict


    def __init__(self, evaluator, scan_map, mass_shift_map, solution_packer, evaluation_args=None):
        if evaluation_args is None:
            evaluation_args = dict()
        self.evaluator = evaluator
        self.scan_map = scan_map
        self.mass_shift_map = mass_shift_map
        self.solution_packer = solution_packer
        self.solution_map = dict()
        self.evaluation_args = evaluation_args
        try:
            self.construct_cache_subgroups = self.evaluator.construct_cache_subgroups
        except AttributeError:
            self.log("Failed to acquire construct_cache_subgroups")
        try:
            self.create_evaluation_context = self.evaluator.create_evaluation_context
        except AttributeError:
            self.log("Failed to acquire create_evaluation_context")

    def evaluate(self, scan, structure, *args, **kwargs):
        return self.evaluator.evaluate(scan, structure, *args, **kwargs)

    def pack_output(self, target):
        package = (target.id, self.solution_map)
        self.solution_map = dict()
        return package

    def process(self, hit_map: Mapping[Any, Any],
                hit_to_scan_map: Mapping[Any, List[str]],
                scan_hit_type_map: Mapping[Tuple[str, Any], MassShift],
                hit_group_map: Mapping[int, List[Any]] = None) -> Mapping[str, List]:
        deque_builder = TaskDeque()
        deque_builder(hit_map, hit_to_scan_map,
                      scan_hit_type_map, hit_group_map)
        i = 0
        has_groups = bool(hit_group_map)
        if not has_groups:
            n = len(hit_to_scan_map)
            for target, scan_spec in deque_builder:
                i += 1
                if i % self.log_process_cycle == 0:
                    self.log("... %0.2f%% of Hits Searched (%d/%d)" %
                            (i * 100. / n, i, n))
                target_id, result = self.handle_item(target, scan_spec)
                target = hit_map[target_id]
                yield (target, result)
        else:
            n = len(hit_group_map)
            for work_order in deque_builder:
                i += 1
                if i % self.log_process_cycle == 0:
                    self.log("... %0.2f%% of Groups Searched (%d/%d)" %
                             (i * 100. / n, i, n))
                for target_id, result in self.handle_group(work_order):
                    target = hit_map[target_id]
                    yield (target, result)


class SequentialIdentificationProcessor(TaskBase):
    evaluation_method: SpectrumEvaluatorBase
    evaluation_args: dict

    mass_shift_map: Mapping[str, MassShift]
    structure_map: Mapping[Any, Any]
    scan_map: Mapping[str, ProcessedScan]

    solution_handler: 'SolutionHandler'
    solution_handler_type: Type['SolutionHandler']


    def __init__(self, evaluator, mass_shift_map, evaluation_args=None, solution_handler_type=None):
        if evaluation_args is None:
            evaluation_args = dict()
        if solution_handler_type is None:
            solution_handler_type = SolutionHandler
        self.evaluation_method = evaluator
        self.evaluation_args = evaluation_args
        self.mass_shift_map = mass_shift_map
        self.solution_handler_type = solution_handler_type
        self.solution_handler = self.solution_handler_type({}, {}, self.mass_shift_map)
        self.structure_map = None
        self.scan_map = None

    def _make_evaluator(self, **kwargs):
        evaluator = LocalSpectrumEvaluator(
            self.evaluation_method,
            self.scan_map,
            self.mass_shift_map,
            self.solution_handler.packer,
            self.evaluation_args)
        return evaluator

    def process(self, scan_map: Mapping[str, ProcessedScan],
                hit_map: Mapping[Any, Any],
                hit_to_scan_map: Mapping[Any, List[str]],
                scan_hit_type_map: Mapping[Tuple[str, Any], MassShift],
                hit_group_map: Mapping[int, List[Any]] = None) -> Mapping[str, List]:
        self.structure_map = hit_map
        start_time = time.time()
        self.scan_map = self.solution_handler.scan_map = scan_map
        evaluator = self._make_evaluator()
        # self.log("... Searching Hits (%d:%d)" % (
        #     len(hit_to_scan_map),
        #     sum(map(len, hit_to_scan_map.values()))))
        for target, score_map in evaluator.process(hit_map, hit_to_scan_map, scan_hit_type_map, hit_group_map):
            self.store_result(target, score_map)
        end_time = time.time()
        elapsed = end_time - start_time
        self.log("... Identification Completed (%0.2f sec.): %d Solutions" %
                 (elapsed, self.solution_handler.counter, ))
        return self.scan_solution_map

    @property
    def scan_solution_map(self):
        return self.solution_handler.scan_solution_map

    def store_result(self, target, score_map):
        """Save the spectrum match scores for ``target`` against the
        set of matched scans

        Parameters
        ----------
        target : object
            The structure that was matched
        score_map : dict
            Maps (scan id, mass shift name) to score
        """
        self.solution_handler(target, score_map)


class SolutionHandler(TaskBase):
    def __init__(self, scan_solution_map, scan_map, mass_shift_map, packer=None):
        if packer is None:
            packer = SolutionPacker()
        self.scan_solution_map = defaultdict(list, (scan_solution_map or {}))
        self.scan_map = scan_map
        self.mass_shift_map = mass_shift_map
        self.packer = packer
        self.counter = 0

    def _make_spectrum_match(self, scan_id, target, score, shift_type):
        return SpectrumMatch(
            self.scan_map[scan_id], target, score,
            mass_shift=self.mass_shift_map[shift_type])

    def store_result(self, target, score_map):
        """Save the spectrum match scores for ``target`` against the
        set of matched scans

        Parameters
        ----------
        target : object
            The structure that was matched
        score_map : dict
            Maps (scan id, mass shift name) to score
        """
        self.counter += 1
        for hit_spec, result_pack in score_map.items():
            scan_id, shift_type = hit_spec
            score = self.packer.unpack(result_pack)
            psm = self._make_spectrum_match(scan_id, target, score, shift_type)
            self.scan_solution_map[scan_id].append(psm)

    def __call__(self, target, score_map):
        return self.store_result(target, score_map)


class SolutionPacker(object):
    def __call__(self, spectrum_match):
        return spectrum_match.score

    def unpack(self, package):
        return package


class MultiScoreSolutionHandler(SolutionHandler):
    def __init__(self, scan_solution_map, scan_map, mass_shift_map, packer=None, spectrum_match_type=None):
        if spectrum_match_type is None:
            spectrum_match_type = MultiScoreSpectrumMatch
        if packer is None:
            packer = MultiScoreSolutionPacker(score_set_tp=spectrum_match_type.score_set_type)
        self.spectrum_match_type = spectrum_match_type
        super(MultiScoreSolutionHandler, self).__init__(
            scan_solution_map, scan_map, mass_shift_map, packer)

    def _make_spectrum_match(self, scan_id, target, score, shift_type):
        return self.spectrum_match_type(
            self.scan_map[scan_id], target, score,
            mass_shift=self.mass_shift_map[shift_type])


class MultiScoreSolutionPacker(object):
    def __init__(self, score_set_tp=None):
        if score_set_tp is None:
            score_set_tp = ScoreSet
        self.score_set_tp = score_set_tp

    def __call__(self, spectrum_match):
        return self.score_set_tp.from_spectrum_matcher(spectrum_match).pack()

    def unpack(self, package):
        return self.score_set_tp.unpack(package)



import itertools
from collections import Counter, defaultdict

from glypy.structure.glycan_composition import FrozenGlycanComposition
from glypy.composition import formula, Composition

from glycresoft.serialize.hypothesis.glycan import (
    GlycanComposition, GlycanCombination, GlycanCombinationGlycanComposition)

from glycresoft.serialize import DatabaseBoundOperation
from glycresoft.task import TaskBase


class GlycanCompositionRecord(object):
    def __init__(self, db_glycan_composition):
        self.id = db_glycan_composition.id
        self.composition = db_glycan_composition.convert()
        self._str = str(self.composition)
        self._hash = hash(self._str)
        self.mass = self.composition.mass()
        self.elemental_composition = self.composition.total_composition()

    def __repr__(self):
        return "GlycanCompositionRecord(%d, %s)" % (self.id, self.composition)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        return self._str == other._str


class GlycanCombinationSpecification(object):
    def __init__(self, class_combinations=None, default=1):
        if class_combinations is None:
            class_combinations = defaultdict(lambda: self.default)
        self.default = default
        self.class_combinations = class_combinations

    def __getitem__(self, class_spec):
        return self.class_combinations[class_spec]


def merge_compositions_frozen(composition_list):
    """Given a list of monosaccharide packed strings,
    sum the monosaccharide counts across lists and return
    the merged string

    Parameters
    ----------
    composition_list : list of GlycanCompositionRecord

    Returns
    -------
    FrozenGlycanComposition
    """
    first = FrozenGlycanComposition()
    for comp_rec in composition_list:
        first += comp_rec.composition
    return first


class GlycanCombinationBuilder(object):
    def __init__(self, glycan_compositions, max_size=1):
        self.glycan_compositions = list(map(GlycanCompositionRecord, glycan_compositions))
        self.max_size = max_size

    def combinate(self, n=1):
        j = 0
        for comb_compositions in itertools.combinations_with_replacement(self.glycan_compositions, n):
            j += 1
            counts = Counter(comb_compositions)
            merged = merge_compositions_frozen(comb_compositions)
            composition = str(merged)
            mass = sum(c.mass for c in comb_compositions)
            elemental_composition = Composition()
            for c in comb_compositions:
                elemental_composition += c.elemental_composition
            inst = GlycanCombination(
                count=n,
                calculated_mass=mass,
                composition=composition,
                formula=formula(elemental_composition))
            yield inst, counts

    def combinate_all(self):
        for i in range(1, self.max_size + 1):
            for data in self.combinate(i):
                yield data


class GlycanCombinationSerializer(DatabaseBoundOperation, TaskBase):
    def __init__(self, connection, source_hypothesis_id, target_hypothesis_id, max_size=1):
        DatabaseBoundOperation.__init__(self, connection)
        self.source_hypothesis_id = source_hypothesis_id
        self.target_hypothesis_id = target_hypothesis_id
        self.max_size = max_size
        self.total_count = 0

    def load_compositions(self):
        return self.query(GlycanComposition).filter(
            GlycanComposition.hypothesis_id == self.source_hypothesis_id).all()

    def generate(self):
        compositions = self.load_compositions()
        combinator = GlycanCombinationBuilder(compositions, self.max_size)

        relation_acc = []
        i = 0
        j = 0
        hypothesis_id = self.target_hypothesis_id
        self.log("... Building combinations for Hypothesis %d" % hypothesis_id)
        for comb, counts in combinator.combinate_all():
            comb.hypothesis_id = hypothesis_id
            self.session.add(comb)
            self.session.flush()

            for key, count in counts.items():
                relation_acc.append({
                    "glycan_id": key.id,
                    "count": count,
                    "combination_id": comb.id
                })

            i += 1
            j += 1
            self.total_count += 1
            if i > 50000:
                self.log(f"...... {j} combinations created")
                self.session.execute(
                    GlycanCombinationGlycanComposition.insert(), relation_acc)
                i = 0
                relation_acc = []
        if relation_acc:
            self.session.execute(
                GlycanCombinationGlycanComposition.insert(), relation_acc)
        self.session.commit()

    def run(self):
        self.generate()
        total = self.session.query(
            GlycanCombination).filter(
            GlycanCombination.hypothesis_id == self.target_hypothesis_id
        ).count()
        self.log(f"{total} Glycan Combinations Constructed.")



import unittest

import glycopeptidepy

from glypy.structure.glycan_composition import HashableGlycanComposition

from glycresoft.chromatogram_tree.mass_shift import Unmodified, Ammonium
from glycresoft.scoring.elution_time_grouping.structure import ChromatogramProxy, GlycopeptideChromatogramProxy
from glycresoft.scoring.elution_time_grouping.reviser import IsotopeRule


class TestChromatogramProxy(unittest.TestCase):
    def _create_instance(self) -> ChromatogramProxy:
        gc = HashableGlycanComposition(Hex=5, HexNAc=4, NeuAc=2, Fuc=1)
        return ChromatogramProxy(gc.mass(), 25.0, 1e6, gc, None, [Unmodified, Ammonium])

    def test_apex_time(self):
        proxy = self._create_instance()
        self.assertAlmostEqual(proxy.apex_time, 25.0, 3)

    def test_weight(self):
        proxy = self._create_instance()
        assert proxy.weight == 1.0

    def test_copy(self):
        proxy = self._create_instance()
        assert proxy == proxy.copy()

    def test_shift_glycan(self):
        proxy = self._create_instance()
        new = proxy.shift_glycan_composition(HashableGlycanComposition(Fuc=2, NeuAc=-1))
        assert new.glycan_composition == HashableGlycanComposition(
            Hex=5, HexNAc=4, NeuAc=1, Fuc=3)


class TestGlycopeptideChromatogramProxy(TestChromatogramProxy):
    def _create_instance(self) -> GlycopeptideChromatogramProxy:
        gc = HashableGlycanComposition(Hex=5, HexNAc=4, NeuAc=2, Fuc=1)
        glycopeptide = glycopeptidepy.parse("NEEYN(N-Glycosylation)K" + str(gc))
        inst = GlycopeptideChromatogramProxy(glycopeptide.total_mass, 25.0, 1e6, gc, None, [Unmodified, Ammonium], structure=glycopeptide)
        assert inst.structure == glycopeptide
        return inst

    def test_shift_glycan(self):
        proxy = self._create_instance()
        new = proxy.shift_glycan_composition(
            HashableGlycanComposition(Fuc=2, NeuAc=-1))
        assert new.glycan_composition == HashableGlycanComposition(
            Hex=5, HexNAc=4, NeuAc=1, Fuc=3)
        assert new.structure == glycopeptidepy.parse(
            "NEEYN(N-Glycosylation)K" + str(new.glycan_composition))
        assert new.kwargs['original_structure'] == str(proxy.structure)


class TestRevisionRule(unittest.TestCase):
    def _create_instance(self) -> GlycopeptideChromatogramProxy:
        gc = HashableGlycanComposition(Hex=5, HexNAc=4, NeuAc=1, Fuc=3)
        glycopeptide = glycopeptidepy.parse("NEEYN(N-Glycosylation)K" + str(gc))
        inst = GlycopeptideChromatogramProxy(glycopeptide.total_mass, 25.0, 1e6, gc, None, [Unmodified, Ammonium], structure=glycopeptide)
        assert inst.structure == glycopeptide
        return inst

    def test_apply(self):
        proxy = self._create_instance()
        rule = IsotopeRule.clone()
        new = rule(proxy)
        assert new.glycan_composition == HashableGlycanComposition(
            Hex=5, HexNAc=4, NeuAc=2, Fuc=1)
        assert new.structure == glycopeptidepy.parse(
            "NEEYN(N-Glycosylation)K" + str(new.glycan_composition))
        assert new.kwargs['original_structure'] == str(proxy.structure)

    def test_apply_with_cache(self):
        proxy = self._create_instance()
        rule = IsotopeRule.with_cache()
        new = rule(proxy)
        assert new.glycan_composition == HashableGlycanComposition(
            Hex=5, HexNAc=4, NeuAc=2, Fuc=1)
        assert new.structure == glycopeptidepy.parse(
            "NEEYN(N-Glycosylation)K" + str(new.glycan_composition))
        assert new.kwargs['original_structure'] == str(proxy.structure)

        new2 = rule(proxy)
        assert new2.glycan_composition == HashableGlycanComposition(
            Hex=5, HexNAc=4, NeuAc=2, Fuc=1)
        assert new2.structure == glycopeptidepy.parse(
            "NEEYN(N-Glycosylation)K" + str(new.glycan_composition))
        assert new2.kwargs['original_structure'] == str(proxy.structure)

        assert new2 is not new
        assert new2.structure is not new.structure



import re

from uuid import uuid4
from collections import OrderedDict

from glycresoft.serialize.hypothesis import GlycanHypothesis
from glycresoft.serialize.hypothesis.glycan import (
    GlycanComposition as DBGlycanComposition, GlycanClass,
    GlycanCompositionToClass, GlycanTypes)

from glycresoft.serialize import DatabaseBoundOperation, func

from glycresoft.database.builder.base import HypothesisSerializerBase

from glypy.composition import composition_transform, formula
from glypy.structure import glycan_composition
from glypy.io import byonic, glyconnect
from glypy.utils import opener
from glypy import ReducedEnd


class TextFileGlycanCompositionLoader(object):
    def __init__(self, file_object, dialect=None):
        if dialect is None:
            dialect = dispatching_dialect
        self.file_object = file_object
        self.current_line = 0
        self.dialect = dialect

    def _partition_line(self, line):
        return self.dialect.parse_line(line)

    def _process_line(self):
        line = next(self.file_object)
        line = line.strip()
        self.current_line += 1
        try:
            while line == '':
                line = next(self.file_object)
                line = line.strip()
                self.current_line += 1
            tokens = self._partition_line(line)
            if len(tokens) == 1:
                composition, structure_classes = tokens[0], ()
            elif len(tokens) == 2:
                composition, structure_classes = tokens
                structure_classes = [structure_classes]
            else:
                composition = tokens[0]
                structure_classes = tokens[1:]
            gc = self.dialect.parse_composition(composition)
        except StopIteration:
            raise
        except Exception as e:
            raise Exception("Parsing Error %r occurred at line %d" % (e, self.current_line))
        return gc, structure_classes

    def __next__(self):
        return self._process_line()

    next = __next__

    def __iter__(self):
        return self

    def close(self):
        self.file_object.close()

class GlycanCompositionDialectError(Exception):
    pass


class GlycanCompositionDialect(object):
    name = 'iupaclite'

    def parse_line(self, line):
        parts, offset, n = self.split_line(line)
        # line is just glycan composition
        if offset is None:
            return parts
        i = 0
        while (offset + i) < n:
            c = line[offset + i]
            if c.isspace():
                break
            i += 1
        parts.append(line[:offset + i])
        parts.extend(t for t in re.split(
            r"(?:\t|\s{2,})", line[offset + i:]) if t)
        return parts

    def split_line(self, line):
        n = len(line)
        try:
            brace_close = line.index("}")
        except ValueError:
            raise GlycanCompositionDialectError()
        parts = []
        # line is just glycan composition
        if brace_close == n - 1:
            parts.append(line)
            return parts, None, n
        offset = brace_close + 1
        return parts, offset, n

    def validate_composition(self, composition):
        if not composition:
            raise Exception()
        for mono in composition:
            if mono.mass() == 0.0:
                raise ValueError(
                    "Invalid glycan composition element %r, no known mass")
        return composition

    def parse_composition(self, composition):
        result = glycan_composition.GlycanComposition.parse(composition)
        result = self.validate_composition(result)
        return result


class SimpleGlycanCompositionDialect(GlycanCompositionDialect):

    def split_line(self, line):
        n = len(line)
        sep = re.search(r"(\t+|\s{2,})", line)
        if sep is None:
            raise GlycanCompositionDialectError()
        else:
            junction = sep.start()
        parts = []
        offset = junction
        return parts, offset, n

    def parse_composition(self, composition):
        result = super(SimpleGlycanCompositionDialect, self).parse_composition(composition)
        return result


class ByonicGlycanCompositionDialect(SimpleGlycanCompositionDialect):
    name = 'byonic'

    def parse_composition(self, composition):
        result = byonic.loads(composition).thaw()
        result = self.validate_composition(result)
        return result


class GlyconnectGlycanCompositionDialect(SimpleGlycanCompositionDialect):
    name = 'glyconnect'

    def parse_composition(self, composition):
        result = glyconnect.loads(composition).thaw()
        result = self.validate_composition(result)
        return result


dialects = OrderedDict([
    ("iupaclite", GlycanCompositionDialect()),
    ("byonic", ByonicGlycanCompositionDialect()),
    ("glyconnect", GlyconnectGlycanCompositionDialect())
])


class DispatchingGlycanCompositionDialect(object):
    def __init__(self, dialects):
        self.dialects = dialects

    def parse_line(self, line):
        for name, dialect in self.dialects.items():
            try:
                return dialect.parse_line(line)
            except GlycanCompositionDialectError:
                continue
        raise GlycanCompositionDialectError()

    def parse_composition(self, composition):
        for name, dialect in self.dialects.items():
            try:
                return dialect.parse_composition(composition)
            except Exception as err:
                continue
        raise GlycanCompositionDialectError()


dispatching_dialect = DispatchingGlycanCompositionDialect(dialects)


_default_label_map = {
    "n glycan": GlycanTypes.n_glycan,
    "n linked": GlycanTypes.n_glycan,
    "o glycan": GlycanTypes.o_glycan,
    "o linked": GlycanTypes.o_glycan,
    "gag linker": GlycanTypes.gag_linker
}


def normalize_lookup(string):
    normalized_string = string.lower().replace("-", " ").replace(
        "\"", '').replace("'", '').strip().rstrip()
    if normalized_string in _default_label_map:
        return _default_label_map[normalized_string]
    else:
        return string


class GlycanClassLoader(object):
    def __init__(self, session):
        self.session = session
        self.store = dict()

    def get(self, name):
        name = normalize_lookup(name)
        try:
            return self.store[name]
        except KeyError:
            glycan_class = self.session.query(GlycanClass).filter(GlycanClass.name == name).first()
            if glycan_class is None:
                glycan_class = GlycanClass(name=name)
                self.session.add(glycan_class)
                self.session.flush()
            self.store[name] = glycan_class
            return glycan_class

    def __getitem__(self, name):
        return self.get(name)


named_reductions = {
    'reduced': 'H2',
    'deuteroreduced': 'HH[2]',
    '2ab': "C7H8N2",
    '2aa': "C7H7NO",
    'apts': "C16H11N1O8S3",
}


named_derivatizations = {
    "permethylated": "methyl",
    "peracetylated": "acetyl"
}


class GlycanTransformer(object):
    def __init__(self, glycan_source, reduction=None, derivatization=None):
        self.glycan_source = glycan_source
        self.reduction = reduction
        self.derivatization = derivatization

        if reduction is not None and isinstance(reduction, str):
            self.reduction = ReducedEnd(reduction)

    def _process_composition(self):
        gc, structure_classes = next(self.glycan_source)
        if self.reduction is not None and gc.reducing_end is None:
            gc.reducing_end = self.reduction.clone()
        if self.derivatization is not None:
            gc = composition_transform.derivatize(gc, self.derivatization)
        return gc, structure_classes

    def __next__(self):
        return self._process_composition()

    next = __next__

    def __iter__(self):
        return self


class GlycanHypothesisSerializerBase(DatabaseBoundOperation, HypothesisSerializerBase):
    def __init__(self, database_connection, hypothesis_name=None, uuid=None):
        if uuid is None:
            uuid = str(uuid4().hex)
        DatabaseBoundOperation.__init__(self, database_connection)
        self._hypothesis_name = hypothesis_name
        self._hypothesis_id = None
        self._hypothesis = None
        self._structure_class_loader = None
        self.uuid = uuid

    def make_structure_class_loader(self):
        return GlycanClassLoader(self.session)

    @property
    def structure_class_loader(self):
        if self._structure_class_loader is None:
            self._structure_class_loader = self.make_structure_class_loader()
        return self._structure_class_loader

    def _construct_hypothesis(self):
        if self._hypothesis_name is None:
            self._hypothesis_name = self._make_name()
        self._hypothesis = GlycanHypothesis(name=self._hypothesis_name, uuid=self.uuid)

        self.session.add(self._hypothesis)
        self.session.commit()

        self._hypothesis_id = self._hypothesis.id
        self._hypothesis_name = self._hypothesis.name

    def _make_name(self):
        return "GlycanHypothesis-" + self.uuid


class TextFileGlycanHypothesisSerializer(GlycanHypothesisSerializerBase):
    def __init__(self, glycan_text_file, database_connection, reduction=None, derivatization=None,
                 hypothesis_name=None):
        GlycanHypothesisSerializerBase.__init__(self, database_connection, hypothesis_name)

        self.glycan_file = glycan_text_file
        self.reduction = reduction
        self.derivatization = derivatization

        self.loader = None
        self.transformer = None

    def _make_name(self):
        return "TextFileGlycanHypothesis-" + self.uuid

    def make_pipeline(self):
        self.loader = TextFileGlycanCompositionLoader(opener(self.glycan_file, 'r'))
        self.transformer = GlycanTransformer(self.loader, self.reduction, self.derivatization)

    def run(self):
        self.make_pipeline()
        structure_class_lookup = self.structure_class_loader
        self.log("Loading Glycan Compositions from Stream for %r" % self.hypothesis)

        acc = []
        counter = 0
        for composition, structure_classes in self.transformer:
            mass = composition.mass()
            composition_string = composition.serialize()
            formula_string = formula(composition.total_composition())
            inst = DBGlycanComposition(
                calculated_mass=mass, formula=formula_string,
                composition=composition_string,
                hypothesis_id=self.hypothesis_id)
            self.session.add(inst)
            self.session.flush()
            counter += 1
            for structure_class in structure_classes:
                structure_class = structure_class_lookup[structure_class]
                acc.append(dict(glycan_id=inst.id, class_id=structure_class.id))
                if len(acc) % 100 == 0:
                    self.session.execute(GlycanCompositionToClass.insert(), acc)
                    acc = []
        if acc:
            self.session.execute(GlycanCompositionToClass.insert(), acc)
            acc = []
        self.session.commit()
        self.loader.close()
        self.log("Generated %d glycan compositions" % counter)


class GlycanCompositionHypothesisMerger(GlycanHypothesisSerializerBase):
    def __init__(self, database_connection, source_hypothesis_ids, hypothesis_name, uuid=None):
        GlycanHypothesisSerializerBase.__init__(
            self, database_connection, hypothesis_name, uuid=uuid)
        self.source_hypothesis_ids = source_hypothesis_ids

    def stream_from_hypotheses(self, connection, hypothesis_id):
        self.log("Streaming from %s for hypothesis %d" % (connection, hypothesis_id))
        connection = DatabaseBoundOperation(connection)
        session = connection.session()
        for db_composition in session.query(DBGlycanComposition).filter(
                DBGlycanComposition.hypothesis_id == hypothesis_id):
            structure_classes = list(db_composition.structure_classes)
            if len(structure_classes) > 0:
                yield db_composition, [sc.name for sc in db_composition.structure_classes]
            else:
                yield db_composition, [None]

    def extract_iterative(self):
        for connection, hypothesis_id in self.source_hypothesis_ids:
            for pair in self.stream_from_hypotheses(connection, hypothesis_id):
                yield pair

    def run(self):
        structure_class_lookup = self.structure_class_loader
        acc = []
        seen = set()
        composition_cache = dict()
        counter = 0
        for db_composition, structure_classes in self.extract_iterative():
            counter += 1
            composition_string = db_composition.composition
            novel_combinations = set()
            for structure_class in structure_classes:
                if (composition_string, structure_class) not in seen:
                    novel_combinations.add(structure_class)
            if len(novel_combinations) == 0:
                continue
            try:
                inst = composition_cache[composition_string]
            except KeyError:
                mass = db_composition.calculated_mass
                formula_string = db_composition.formula
                inst = DBGlycanComposition(
                    calculated_mass=mass, formula=formula_string,
                    composition=composition_string,
                    hypothesis_id=self.hypothesis_id)
                self.session.add(inst)
                self.session.flush()
                composition_cache[composition_string] = inst
            for structure_class in novel_combinations:
                seen.add((composition_string, structure_class))
                if structure_class is None:
                    continue
                structure_class = structure_class_lookup[structure_class]
                acc.append(dict(glycan_id=inst.id, class_id=structure_class.id))
                if (len(acc) % 10000) == 0:
                    self.session.execute(GlycanCompositionToClass.insert(), acc)
                    acc = []
        if acc:
            self.session.execute(GlycanCompositionToClass.insert(), acc)
            acc = []
        self.session.commit()
        assert len(seen) > 0



import re
import textwrap
from io import BytesIO, TextIOWrapper

from collections import OrderedDict
from typing import Union

from glycopeptidepy.io.fasta import FastaFileParser, FastaFileWriter, PEFFReader
from glycopeptidepy.io.annotation import from_peff

from glycresoft.serialize.hypothesis.peptide import Protein
from .sequence_tree import SuffixTree


class ProteinFastaFileParser(FastaFileParser):
    def __init__(self, *args, **kwargs):
        super(ProteinFastaFileParser, self).__init__(*args, **kwargs)

    def process_result(self, d):
        p = Protein(name=str(d['name']), protein_sequence=d['sequence'])
        return p


class PEFFProteinFastaFileParser(PEFFReader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def process_result(self, d):
        p = Protein(name=str(d['name']), protein_sequence=d['sequence'])
        p.annotations = from_peff(d['name'])
        return p


def open_fasta(path: str):
    if path.endswith(".peff") or path.endswith(".peff.gz"):
        return PEFFProteinFastaFileParser(path)
    else:
        return ProteinFastaFileParser(path)


class ProteinFastaFileWriter(FastaFileWriter):

    def write(self, protein):
        defline = ''.join(
            [">", protein.name, " ", str(protein.glycosylation_sites)])
        seq = '\n'.join(textwrap.wrap(protein.protein_sequence, 80))
        super(ProteinFastaFileWriter, self).write(defline, seq)

    def writelines(self, iterable):
        for protein in iterable:
            self.write(protein)


class SequenceLabel(object):

    def __init__(self, label, sequence):
        self.label = label
        self.sequence = sequence

    def __getitem__(self, i):
        if isinstance(i, int):
            return self.label[i]
        else:
            return SequenceLabel(self.label[i], self.sequence)

    def __len__(self):
        return len(self.label)

    def __iter__(self):
        return iter(self.label)

    def __repr__(self):
        return "SequenceLabel(%s, %r)" % (self.label, self.sequence)


class FastaIndex(object):
    def __init__(self, path, block_size=1000000, encoding='utf-8'):
        self.path = path
        self.block_size = block_size
        self.index = OrderedDict()
        self.encoding = encoding

        self.build_index()

    def _chunk_iterator(self):
        delim = b"\n>"
        read_size = self.block_size
        with open(self.path, 'rb') as f:
            buff = f.read(read_size)
            started_with_with_delim = buff.startswith(delim)
            parts = buff.split(delim)
            tail = parts[-1]
            front = parts[:-1]
            i = 0
            for part in front:
                i += 1
                if part == b'':
                    continue
                if i == 1:
                    if started_with_with_delim:
                        yield delim + part
                    else:
                        yield part
                else:
                    yield delim + part
            running = True
            while running:
                buff = f.read(read_size)
                if len(buff) == 0:
                    running = False
                    buff = tail
                else:
                    buff = tail + buff
                parts = buff.split(delim)
                tail = parts[-1]
                front = parts[:-1]
                for part in front:
                    yield delim + part
            yield delim + tail

    def _generate_offsets(self):
        i = 0
        defline_pattern = re.compile(br"\s*>([^\n]+)\n")
        for line in self._chunk_iterator():
            match = defline_pattern.match(line)
            if match:
                yield i, match.group(1)
            i += len(line)
        yield i, None

    def build_index(self):
        index = OrderedDict()
        g = self._generate_offsets()
        last_offset = 0
        last_defline = None
        for offset, defline in g:
            if last_defline is not None:
                index[last_defline] = (last_offset, offset)
            last_defline = defline
            last_offset = offset
        assert last_defline is None
        self.index = index

    def _ensure_bytes(self, string):
        if isinstance(string, bytes):
            return string
        try:
            return string.encode(self.encoding)
        except (AttributeError, UnicodeEncodeError):
            raise TypeError("{!r} could not be encoded".format(string))

    def _offset_of(self, key):
        key = self._ensure_bytes(key)
        return self.index[key]

    def __len__(self):
        return len(self.index)

    def keys(self):
        return self.index.keys()

    def __iter__(self):
        return iter(self.index)

    def fetch_sequence(self, key):
        start, end = self._offset_of(key)
        with open(self.path, 'rb') as f:
            f.seek(start)
            data = f.read(end - start)
            buff = TextIOWrapper(BytesIO(data))
            return next(ProteinFastaFileParser(buff))

    def __getitem__(self, key):
        return self.fetch_sequence(key)


class DeflineSuffix(object):
    def __init__(self, label, original):
        self.label = label
        self.original = original

    def __getitem__(self, i):
        if isinstance(i, int):
            return self.label[i]
        else:
            return DeflineSuffix(self.label[i], self.original)

    def __len__(self):
        return len(self.label)

    def __iter__(self):
        return iter(self.label)

    def __repr__(self):
        return "DeflineSuffix(%s, %r)" % (self.label, self.original)


class FastaProteinSequenceResolver(object):
    def __init__(self, path, encoding='utf-8', strategy='exact'):
        self.path = path
        self.encoding = encoding
        self.index = FastaIndex(self.path, encoding=self.encoding)
        self.resolver = None
        self.strategy = strategy

        self._build_resolver(strategy)

    def _build_resolver(self, strategy='exact'):
        if strategy == 'suffix':
            keys = self.index.keys()
            resolver = SuffixTree()
            i = 0
            for key in keys:
                i += 1
                label = DeflineSuffix(key.decode(self.encoding).split(" ")[0], key)
                resolver.add_ngram(label)
            self.resolver = resolver
        elif strategy == 'exact':
            keys = self.index.keys()
            resolver = dict()
            i = 0
            for key in keys:
                i += 1
                label = DeflineSuffix(key.decode(self.encoding).split(" ")[0], key)
                resolver[label.label] = label
            self.resolver = resolver

    def find(self, name):
        if self.strategy == 'suffix':
            labels = self.resolver.subsequences_of(name)
            sequences = [
                self.index[label.original] for label in labels]
        else:
            try:
                label = self.resolver[name]
                return [self.index[label.original]]
            except KeyError:
                return []
        return sequences

    def __getitem__(self, name):
        return self.find(name)


class ProteinSequenceListResolver(object):

    @classmethod
    def build_from_fasta(cls, fasta):
        return cls(ProteinFastaFileParser(fasta))

    def __init__(self, protein_list):
        protein_list = [SequenceLabel(str(prot.name), prot)
                        for prot in protein_list]
        self.resolver = SuffixTree()
        for prot in protein_list:
            self.resolver.add_ngram(prot)

    def find(self, name):
        return [label_seq.sequence for label_seq in self.resolver.subsequences_of(name)]

    def __getitem__(self, name):
        return self.find(name)



'''Implementation for journal file reading and writing to serialize
glycopeptide spectrum match information to disk during processing.
'''
import os
import io
import csv
import json
import gzip

from multiprocessing import Event, JoinableQueue

from collections import defaultdict
from operator import attrgetter
from typing import Any, Callable, DefaultDict, Dict, List, Optional, Set, Type, Union, TYPE_CHECKING, BinaryIO


import numpy as np

from glycresoft.chromatogram_tree.mass_shift import MassShiftBase

try:
    from pyzstd import ZstdFile
except ImportError:
    ZstdFile = None

from glycopeptidepy.utils import collectiontools

from ms_deisotope.output import ProcessedMSFileLoader

from glycresoft.task import TaskBase, TaskExecutionSequence, Empty

from glycresoft.structure.structure_loader import (
    PeptideProteinRelation, LazyGlycopeptide)
from glycresoft.structure.scan import ScanInformation, ScanInformationLoader
from glycresoft.structure.lru import LRUMapping
from glycresoft.chromatogram_tree import Unmodified

from glycresoft.tandem.ref import SpectrumReference
from glycresoft.tandem.spectrum_match import (
    MultiScoreSpectrumMatch, MultiScoreSpectrumSolutionSet, ScoreSet, FDRSet)

from .search_space import glycopeptide_key_t, StructureClassification

if TYPE_CHECKING:
    from glycresoft.serialize import Analysis, AnalysisDeserializer
    from ms_deisotope.data_source import RandomAccessScanSource


def _zstdopen(path, mode='w'):
    handle = ZstdFile(path, mode=mode.replace('t', 'b'))
    return io.TextIOWrapper(handle, encoding='utf8')


def _zstwrap(fh, mode='w'):
    handle = ZstdFile(fh, mode=mode)
    return io.TextIOWrapper(handle, encoding='utf8')


def _gzopen(path, mode='w'):
    handle = gzip.open(path, mode=mode.replace('t', 'b'))
    return io.TextIOWrapper(handle, encoding='utf8')


def _gzwrap(fh, mode='w'):
    handle = gzip.GzipFile(fileobj=fh, mode=mode)
    return io.TextIOWrapper(handle, encoding='utf8')


if ZstdFile is None:
    _file_opener = _gzopen
    _file_wrapper = _gzwrap
else:
    _file_opener = _zstdopen
    _file_wrapper = _zstwrap


class JournalFileWriter(TaskBase):
    """A task for writing glycopeptide spectrum matches to a TSV-formatted
    journal file. This format is an intermediary result, and will contain many
    random or non-useful matches.

    """

    path: os.PathLike
    handle: io.TextIOWrapper

    include_auxiliary: bool
    include_fdr: bool
    writer: csv.writer
    spectrum_counter: int
    solution_counter: int

    score_columns: List[str]

    def __init__(self, path, include_fdr=False, include_auxiliary=False, score_columns: List[str]=None):
        if score_columns is None:
            score_columns = ScoreSet.field_names()
        self.path = path
        if not hasattr(path, 'write'):
            self.handle = _file_opener(path, 'wb')
        else:
            self.handle = _file_wrapper(self.path, 'w')
        self.include_fdr = include_fdr
        self.include_auxiliary = include_auxiliary
        self.spectrum_counter = 0
        self.solution_counter = 0

        self.score_columns = score_columns

        self.writer = csv.writer(self.handle, delimiter='\t')
        self.write_header()

    def _get_headers(self):
        names = [
            'scan_id',
            'precursor_mass_accuracy',
            'peptide_start',
            'peptide_end',
            'peptide_id',
            'protein_id',
            'hypothesis_id',
            'glycan_combination_id',
            'match_type',
            'site_combination_index',
            'glycosylation_type',
            'glycopeptide_sequence',
            'mass_shift',
        ]
        names.extend(self.score_columns)
        if self.include_fdr:
            names.extend([
                "peptide_q_value",
                "glycan_q_value",
                "glycopeptide_q_value",
                "total_q_value"
            ])
        if self.include_auxiliary:
            names.append("auxiliary")
        return names

    def write_header(self):
        self.writer.writerow(self._get_headers())

    def _prepare_fields(self, psm: MultiScoreSpectrumMatch) -> List[str]:
        error = (psm.target.total_mass - psm.precursor_information.neutral_mass
                ) / psm.precursor_information.neutral_mass
        fields = ([psm.scan_id, error, ] + list(psm.target.id) + [
            psm.target,
            psm.mass_shift.name,
        ])
        fields.extend(psm.score_set.values())
        if self.include_fdr:
            q_value_set = psm.q_value_set
            if q_value_set is None:
                fdr_fields = [
                    1, 1, 1, 1
                ]
            else:
                fdr_fields = [
                    q_value_set.peptide_q_value,
                    q_value_set.glycan_q_value,
                    q_value_set.glycopeptide_q_value,
                    q_value_set.total_q_value
                ]
            fields.extend(fdr_fields)
        if self.include_auxiliary:
            fields.append(json.dumps(psm.get_auxiliary_data(), sort_keys=True))
        fields = [str(f) for f in fields]
        return fields

    def write(self, psm: MultiScoreSpectrumMatch):
        self.solution_counter += 1
        self.writer.writerow(self._prepare_fields(psm))

    def writeall(self, solution_sets: List[MultiScoreSpectrumSolutionSet]):
        for solution_set in solution_sets:
            self.spectrum_counter += 1
            for solution in solution_set:
                self.write(solution)
        self.flush()

    def flush(self):
        self.handle.flush()

    def close(self):
        self.handle.close()


class JournalingConsumer(TaskExecutionSequence):
    journal_file: JournalFileWriter
    in_queue: JoinableQueue
    in_done_event: Event
    done_event: Event

    def __init__(self, journal_file, in_queue, in_done_event):
        self.journal_file = journal_file
        self.in_queue = in_queue
        self.in_done_event = in_done_event
        self.done_event = self._make_event()

    def run(self):
        has_work = True
        while has_work and not self.error_occurred():
            try:
                solutions = self.in_queue.get(True, 5)
                self.journal_file.writeall(solutions)
                # Only log if something changed.
                if solutions:
                    self.debug("... Handled %d spectra with %d solutions so far" % (
                        self.journal_file.spectrum_counter, self.journal_file.solution_counter))
            except Empty:
                if self.in_done_event.is_set():
                    has_work = False
                    break
        self.done_event.set()

    def close_stream(self):
        self.journal_file.close()


def parse_float(value: str) -> float:
    value = float(value)
    if np.isnan(value):
        return 0
    return value


try:
    _parse_float = parse_float
    from glycresoft._c.tandem.tandem_scoring_helpers import parse_float
except ImportError:
    pass


class JournalFileIndex(TaskBase):
    def __init__(self, path):
        self.path = path
        self.scan_ids = set()
        self.glycopeptides = set()
        self.build_index()

    def build_index(self):
        reader = self.open()
        for row in reader:
            self.scan_ids.add(row['scan_id'])
            self.glycopeptides.add(row['glycopeptide_sequence'])

    def has_scan(self, scan_id: str) -> bool:
        return scan_id in self.scan_ids

    def has_glycopeptide(self, glycopeptide: str) -> bool:
        return glycopeptide in self.glycopeptides

    def __repr__(self):
        return f"{self.__class__.__name__}({self.path})"

    def __contains__(self, key: str):
        return self.has_glycopeptide(key) or self.has_scan(key)

    def open(self):
        if not hasattr(self.path, 'read') and not hasattr(self.path, 'open'):
            handle = _file_opener(self.path, 'rt')
        elif hasattr(self.path, 'open'):
            handle = _file_wrapper(self.path.open(raw=True), 'r')
        else:
            handle = _file_wrapper(self.path, 'r')
        reader = csv.DictReader(handle, delimiter='\t')
        return reader

    def collect_for_scan_id(self, scan_id: str):
        reader = self.open()
        for row in reader:
            if scan_id == row['scan_id']:
                yield row


class JournalFileReader(TaskBase):
    path: os.PathLike
    handle: io.TextIOBase
    reader: csv.DictReader

    glycopeptide_cache: LRUMapping
    mass_shift_map: Dict[str, MassShiftBase]
    scan_loader: Optional[ScanInformationLoader]
    include_fdr: bool

    score_set_type: Type
    score_columns: List[str]

    def __init__(self, path, cache_size=2 ** 16, mass_shift_map=None, scan_loader=None,
                 include_fdr=False, score_set_type=None):
        if mass_shift_map is None:
            mass_shift_map = {Unmodified.name: Unmodified}
        else:
            mass_shift_map.setdefault(Unmodified.name, Unmodified)
        if score_set_type is None:
            score_set_type = ScoreSet
        self.path = path
        if not hasattr(path, 'read'):
            self.handle = _file_opener(path, 'rt')
        else:
            self.handle = _file_wrapper(self.path, 'r')
        self.reader = csv.DictReader(self.handle, delimiter='\t')
        self.glycopeptide_cache = LRUMapping(cache_size or 2 ** 12)
        self.mass_shift_map = mass_shift_map
        self.scan_loader = scan_loader
        self.include_fdr = include_fdr

        self.score_set_type = score_set_type
        self.score_columns = score_set_type.field_names()
        for name in ScoreSet.field_names():
            self.score_columns.remove(name)

    def _build_key(self, row) -> glycopeptide_key_t:
        glycopeptide_id_key = glycopeptide_key_t(
            int(row['peptide_start']), int(row['peptide_end']), int(
                row['peptide_id']), int(row['protein_id']),
            int(row['hypothesis_id']), int(row['glycan_combination_id']),
            StructureClassification[row['match_type']],
            int(row['site_combination_index']), row.get('glycosylation_type'))
        return glycopeptide_id_key

    def _build_protein_relation(self, glycopeptide_id_key: glycopeptide_key_t) -> PeptideProteinRelation:
        return PeptideProteinRelation(
            glycopeptide_id_key.start_position, glycopeptide_id_key.end_position,
            glycopeptide_id_key.protein_id, glycopeptide_id_key.hypothesis_id)

    def glycopeptide_from_row(self, row: Dict) -> LazyGlycopeptide:
        glycopeptide_id_key = self._build_key(row)
        if glycopeptide_id_key in self.glycopeptide_cache:
            return self.glycopeptide_cache[glycopeptide_id_key]
        if glycopeptide_id_key.structure_type & StructureClassification.target_peptide_decoy_glycan.value:
            glycopeptide = LazyGlycopeptide(row['glycopeptide_sequence'], glycopeptide_id_key)
        else:
            glycopeptide = LazyGlycopeptide(row['glycopeptide_sequence'], glycopeptide_id_key)
        glycopeptide.protein_relation = self._build_protein_relation(glycopeptide_id_key)
        self.glycopeptide_cache[glycopeptide_id_key] = glycopeptide
        return glycopeptide

    def _build_score_set(self, row) -> ScoreSet:
        score_set_entries = [
            parse_float(row['total_score']),
            parse_float(row['peptide_score']),
            parse_float(row['glycan_score']),
            float(row['glycan_coverage']),
            float(row['stub_glycopeptide_intensity_utilization']),
            float(row['oxonium_ion_intensity_utilization']),
            int(row['n_stub_glycopeptide_matches']),
            float(row.get('peptide_coverage', 0.0)),
            float(row.get('total_signal_utilization', 0.0))
        ]
        for name in self.score_columns:
            score_set_entries.append(parse_float(row.get(name, '0.0')))

        score_set = self.score_set_type(*score_set_entries)
        return score_set

    def _build_fdr_set(self, row) -> FDRSet:
        fdr_set = FDRSet(
            float(row['total_q_value']),
            float(row['peptide_q_value']),
            float(row['glycan_q_value']),
            float(row['glycopeptide_q_value']))
        return fdr_set

    def _make_mass_shift(self, row) -> MassShiftBase:
        mass_shift = self.mass_shift_map[row['mass_shift']]
        return mass_shift

    def _make_scan(self, row) -> Union[SpectrumReference, ScanInformation]:
        if self.scan_loader is None:
            scan = SpectrumReference(row['scan_id'])
        else:
            scan = self.scan_loader.get_scan_by_id(row['scan_id'])
        return scan

    def spectrum_match_from_row(self, row) -> MultiScoreSpectrumMatch:
        glycopeptide = self.glycopeptide_from_row(row)
        scan = self._make_scan(row)
        fdr_set = None
        if self.include_fdr:
            fdr_set = self._build_fdr_set(row)
        score_set = self._build_score_set(row)
        mass_shift = self._make_mass_shift(row)
        match = MultiScoreSpectrumMatch(
            scan, glycopeptide, score_set, mass_shift=mass_shift,
            q_value_set=fdr_set,
            match_type=str(glycopeptide.id.structure_type))
        return match

    def __iter__(self):
        return self

    def __next__(self) -> MultiScoreSpectrumMatch:
        return self.spectrum_match_from_row(next(self.reader))

    def next(self):
        return self.__next__()

    def close(self):
        self.handle.close()


def isclose(a, b, rtol=1e-05, atol=1e-08):
    return abs(a - b) <= atol + rtol * abs(b)


class SolutionSetGrouper(TaskBase):
    """
    Partition multi-score glycopeptide identificatins into groups
    according to either glycopeptide target/decoy classification (:attr:`match_type_groups`)
    or by best match to a given scan and target/decoy classification (:attr:`exclusive_match_groups`)

    """
    # All the matches of all match types for every spectrum in no particular order
    spectrum_matches: List[MultiScoreSpectrumMatch]
    # The set of all scan IDs
    spectrum_ids: Set[str]
    # The set of all scan IDs where a target is the best match
    target_owned_spectrum_ids: Set[str]
    # The spectrum solution sets for each target classification group
    match_type_groups: Dict[StructureClassification, List[MultiScoreSpectrumSolutionSet]]
    # The set of all spectrum matches that are the best scoring match for their scan in a specific
    # match type. This is used for FDR estimation.
    exclusive_match_groups: DefaultDict[StructureClassification,
                                        List[MultiScoreSpectrumMatch]]

    def __init__(self, spectrum_matches):
        self.spectrum_matches = list(spectrum_matches)
        self.spectrum_ids = set()
        self.target_owned_spectrum_ids = set()
        self.match_type_groups = self._collect(self.spectrum_matches)
        self.exclusive_match_groups = self._exclusive()

    def __getitem__(self, key):
        return self.exclusive_match_groups[key]

    def __iter__(self):
        return iter(self.exclusive_match_groups.items())

    def _by_scan_id(self):
        acc = []
        for by_scan in collectiontools.groupby(self.spectrum_matches, lambda x: x.scan.id).values():
            scan = by_scan[0].scan
            self.spectrum_ids.add(scan.scan_id)
            ss = MultiScoreSpectrumSolutionSet(scan, by_scan)
            ss.sort()
            acc.append(ss)
        acc.sort(key=lambda x: x.scan.id)
        return acc

    def _collect(
        self, spectrum_matches: List[MultiScoreSpectrumMatch]
    ) -> Dict[StructureClassification, List[MultiScoreSpectrumSolutionSet]]:

        match_type_getter = attrgetter('match_type')
        groups = collectiontools.groupby(spectrum_matches, match_type_getter)
        by_scan_groups = {}
        for group, members in groups.items():
            acc = []
            for by_scan in collectiontools.groupby(members, lambda x: x.scan.id).values():
                scan = by_scan[0].scan
                self.spectrum_ids.add(scan.scan_id)
                ss = MultiScoreSpectrumSolutionSet(scan, by_scan)
                ss.sort()
                acc.append(ss)
            acc.sort(key=lambda x: x.scan.id)
            by_scan_groups[group] = acc
        return by_scan_groups

    def _exclusive(
        self, score_getter: Callable[[MultiScoreSpectrumMatch], float] = None, min_value: float = 0
    ) -> DefaultDict[StructureClassification, List[MultiScoreSpectrumMatch]]:
        if score_getter is None:
            score_getter = attrgetter('score')
        groups = collectiontools.groupby(
            self.spectrum_matches, lambda x: x.scan.id)
        by_match_type = defaultdict(list)
        for scan_id, members in groups.items():
            top_match = max(members, key=score_getter)
            top_score = score_getter(top_match)
            seen = set()
            for match in members:
                if (isclose(top_score, score_getter(match)) and score_getter(match) > min_value
                    and match.match_type not in seen):
                    seen.add(match.match_type)
                    by_match_type[match.match_type].append(match)
                    if match.match_type == StructureClassification.target_peptide_target_glycan:
                        self.target_owned_spectrum_ids.add(scan_id)
        for _group_label, matches in by_match_type.items():
            matches.sort(key=lambda x: (x.scan.id, score_getter(x)))
        return by_match_type

    @property
    def target_matches(self) -> List[MultiScoreSpectrumMatch]:
        try:
            return list(
                filter(
                    lambda x: x.scan.id in self.target_owned_spectrum_ids,
                    self.match_type_groups[StructureClassification.target_peptide_target_glycan]
                )
            )
            # return self.match_type_groups[StructureClassification.target_peptide_target_glycan]
        except KeyError:
            return []

    @property
    def decoy_matches(self) -> List[MultiScoreSpectrumMatch]:
        try:
            return self.match_type_groups[StructureClassification.decoy_peptide_target_glycan]
        except KeyError:
            return []

    def target_count(self):
        return len(self.target_matches)

    def decoy_count(self):
        return len(self.decoy_matches)


class JournalSetLoader(TaskBase):
    """A helper class to load a list of journal file fragments
    into a single cohesive result, as from a previously compiled
    analysis's bundled journal file shards.
    """

    journal_files: List[Union[BinaryIO, os.PathLike, str]]
    journal_reader_args: Dict[str, Any]
    scan_loader: Union['RandomAccessScanSource', ScanInformationLoader]
    mass_shift_map: Dict[str, MassShiftBase]
    solutions: List[MultiScoreSpectrumSolutionSet]

    @classmethod
    def from_analysis(cls, analysis: Union["Analysis", "AnalysisDeserializer"],
                      scan_loader: Optional['RandomAccessScanSource']=None,
                      stub_wrapping: bool=True,
                      score_set_type: Optional[Type[ScoreSet]]=None,
                      **journal_reader_args):
        from glycresoft.serialize import AnalysisDeserializer
        if isinstance(analysis, AnalysisDeserializer):
            analysis = analysis.analysis
        mass_shift_map = {
            m.name: m for m in analysis.parameters['mass_shifts']}
        if scan_loader is None:
            scan_loader = ProcessedMSFileLoader(
                analysis.parameters['sample_path'])
        if stub_wrapping:
            stub_loader = ScanInformationLoader(scan_loader)
        else:
            stub_loader = scan_loader
        if score_set_type is None:
            score_set_type = analysis.parameters[
                'tandem_scoring_model'].get_score_set_type()
        return cls([f.open(raw=True) for f in analysis.files], stub_loader, mass_shift_map,
                   score_set_type=score_set_type, **journal_reader_args)

    def __init__(self, journal_files, scan_loader, mass_shift_map=None, **journal_reader_args):
        if mass_shift_map is None:
            mass_shift_map = {Unmodified.name: Unmodified}
        self.journal_files = journal_files
        self.journal_reader_args = journal_reader_args
        self.scan_loader = scan_loader
        self.mass_shift_map = mass_shift_map
        self.solutions = []

    def load(self):
        n = len(self.journal_files)
        for i, journal_path in enumerate(self.journal_files, 1):
            self.log("... Reading Journal Shard %s, %d/%d" %
                     (journal_path, i, n))
            self._load_identifications_from_journal(journal_path, self.solutions)
        self.log("Partitioning Spectrum Matches...")
        return SolutionSetGrouper(self.solutions)

    def __iter__(self):
        if not self.solutions:
            self.load()
        return iter(self.solutions)

    def _load_identifications_from_journal(self, journal_path: Union[BinaryIO, os.PathLike, str],
                                           accumulator: Optional[List[MultiScoreSpectrumSolutionSet]]=None):
        if accumulator is None:
            accumulator = []
        reader = enumerate(
            JournalFileReader(
                journal_path,
                scan_loader=self.scan_loader,
                mass_shift_map=self.mass_shift_map,
                **self.journal_reader_args),
            len(accumulator))
        i = float(len(accumulator))
        should_log = False
        for i, sol in reader:
            if i % 100000 == 0:
                should_log = True
            if should_log:
                self.log("... %d Solutions Loaded" % (i, ))
                should_log = False
            accumulator.append(sol)
        return accumulator



import os
from typing import Callable, Union, Protocol, runtime_checkable

from six import string_types as basestring

from sqlalchemy import create_engine, event
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.engine import Connectable

from sqlalchemy import exc
from sqlalchemy.engine import Engine
from sqlalchemy.orm.session import Session
from sqlalchemy.orm import Query

from .base import Base
from .migration import Migration
from .tandem import GlycopeptideSpectrumMatch, GlycopeptideSpectrumMatchScoreSet


ConnectFrom = Union[str,
                    Connectable,
                    'ConnectionRecipe',
                    scoped_session,
                    Session]


@runtime_checkable
class HasSession(Protocol):
    session: Union[Session, scoped_session]


def configure_connection(connection: ConnectFrom, create_tables=True):
    if isinstance(connection, basestring):
        try:
            eng = create_engine(connection)
        except exc.ArgumentError:
            eng = SQLiteConnectionRecipe(connection)()
    elif isinstance(connection, Connectable):
        eng = connection
    elif isinstance(connection, ConnectionRecipe):
        eng = connection()
    elif isinstance(connection, (scoped_session, Session)):
        eng = connection.get_bind()
    elif isinstance(connection, HasSession):
        eng = connection.session.get_bind()
        create_tables = False
    else:
        raise ValueError(
            "Could not determine how to get a database connection from %r" % connection)
    if create_tables:
        Base.metadata.create_all(bind=eng)
        Migration(eng, GlycopeptideSpectrumMatch).run()
        Migration(eng, GlycopeptideSpectrumMatchScoreSet).run()
    return eng


initialize = configure_connection

def _noop_on_connect(connection, connection_record):
    pass


class ConnectionRecipe(object):
    connection_url: str
    connect_args: dict
    on_connect: Callable
    engine_args: dict

    def __init__(self, connection_url: str, connect_args=None, on_connect=None, **engine_args):
        if connect_args is None:
            connect_args = {}
        if on_connect is None:
            on_connect = _noop_on_connect

        self.connection_url = connection_url
        self.connect_args = connect_args
        self.on_connect = on_connect
        self.engine_args = engine_args

    def __call__(self) -> Engine:
        connection = create_engine(
            self.connection_url, connect_args=self.connect_args,
            **self.engine_args)
        event.listens_for(connection, 'connect')(self.on_connect)
        return connection


class SQLiteConnectionRecipe(ConnectionRecipe):
    connect_args = {
        'timeout': 180,
    }

    @staticmethod
    def _configure_connection(dbapi_connection, connection_record):
        # disable pysqlite's emitting of the BEGIN statement entirely.
        # also stops it from emitting COMMIT before any DDL.
        iso_level = dbapi_connection.isolation_level
        dbapi_connection.isolation_level = None
        try:
            dbapi_connection.execute("PRAGMA page_size = 5120;")
            dbapi_connection.execute("PRAGMA cache_size = 12000;")
            dbapi_connection.execute("PRAGMA temp_store = 3;")
            if not int(os.environ.get("NOWAL", '0')):
                dbapi_connection.execute("PRAGMA journal_mode = WAL;")
                dbapi_connection.execute("PRAGMA wal_autocheckpoint = 100;")
            # dbapi_connection.execute("PRAGMA foreign_keys = ON;")
            dbapi_connection.execute("PRAGMA journal_size_limit = 1000000;")
            pass
        except Exception as e:
            print(e)
        dbapi_connection.isolation_level = iso_level

    def __init__(self, connection_url, **engine_args):
        super(SQLiteConnectionRecipe, self).__init__(
            self._construct_url(connection_url), self.connect_args, self._configure_connection)

    def _construct_url(self, path):
        if path.startswith("sqlite://"):
            return path
        else:
            return "sqlite:///%s" % path


class DatabaseBoundOperation(object):
    engine: Engine
    _original_connection: ConnectFrom
    _sessionmaker: scoped_session

    def __init__(self, connection):
        self.engine = self._configure_connection(connection)

        self._original_connection = connection
        self._sessionmaker = scoped_session(
            sessionmaker(bind=self.engine, autoflush=False))

    def bridge(self, other: 'DatabaseBoundOperation'):
        self.engine = other.engine
        self._sessionmaker = other._sessionmaker

    def _configure_connection(self, connection: ConnectFrom):
        eng = configure_connection(connection, create_tables=True)
        return eng

    def __eq__(self, other):
        return str(self.engine) == str(other.engine) and\
            (self.__class__ is other.__class__)

    def __hash__(self):
        return hash(str(self.engine))

    def __ne__(self, other):
        return not (self == other)

    def __reduce__(self):
        return self.__class__, (self._original_connection,)

    @property
    def session(self):
        return self._sessionmaker

    def query(self, *args) -> Query:
        return self.session.query(*args)

    def _analyze_database(self):
        conn = self.engine.connect()
        inner = conn.connection.connection
        cur = inner.cursor()
        cur.execute("analyze;")
        inner.commit()

    def _sqlite_reload_analysis_plan(self):
        conn = self.engine.connect()
        conn.execute("ANALYZE sqlite_master;")

    def _sqlite_checkpoint_wal(self):
        conn = self.engine.connect()
        inner = conn.connection.connection
        cur = inner.cursor()
        cur.execute("PRAGMA wal_checkpoint(SQLITE_CHECKPOINT_RESTART);")

    @property
    def dialect(self):
        return self.engine.dialect

    def is_sqlite(self):
        return self.dialect.name == "sqlite"

    def close(self):
        self.session.remove()
        self.engine.dispose()



from .utils import hypothesis_register
from . import heparin
from . import combinatorial_mammalian_n_linked
from . import glycosaminoglycan_linkers
from . import combinatorial_human_n_linked
from . import biosynthesis_human_n_linked
from . import biosynthesis_mammalian_n_linked
from . import human_mucin_o_linked


__all__ = [
    "hypothesis_register",
    "heparin",
    "combinatorial_mammalian_n_linked",
    "combinatorial_human_n_linked",
    "glycosaminoglycan_linkers",
    "biosynthesis_human_n_linked",
    "biosynthesis_mammalian_n_linked",
    "human_mucin_o_linked",
]



# try:   # pragma: no cover
#     from cStringIO import StringIO
# except:  # pragma: no cover
#     try:
#         from StringIO import StringIO
#     except:
#         from io import StringIO
# from io import StringIO
from io import BytesIO
try:  # pragma: no cover
    from lxml import etree as ET
except ImportError:  # pragma: no cover
    try:
        from xml.etree import cElementTree as ET
    except ImportError:
        from xml.etree import ElementTree as ET


class IDMapper(dict):
    '''
    A dictionary-like container which uses a format-string
    key pattern to generate unique identifiers for each entry

    Key Pattern: '<type-name>-%d'

    Associates each generated id with a dictionary of metadata and
    sets the `gid` of the passed `matplotlib.Artist` to the generated
    id. Only the id and metadata are stored.

    Used to preserve a mapping of metadata to artists for later SVG
    serialization.
    '''

    def __init__(self):
        dict.__init__(self)
        self.counter = 0

    def add(self, key, value, meta):
        label = key % self.counter
        value.set_gid(label)
        self[label] = meta
        self.counter += 1
        return label



import os
import functools

from typing import List, Mapping, Optional, TYPE_CHECKING

from glycresoft.chromatogram_tree.mass_shift import MassShift

from ms_deisotope import isotopic_shift
from ms_deisotope.data_source import ProcessedScan

from glycresoft.task import TaskBase
from glycresoft.chromatogram_tree import Unmodified


from .process_dispatcher import (
    IdentificationProcessDispatcher,
    SolutionHandler,
    MultiScoreSolutionHandler,
    MultiScoreSolutionPacker,
    SequentialIdentificationProcessor)

from .workload import WorkloadManager, DEFAULT_WORKLOAD_MAX
from .spectrum_match import (
    MultiScoreSpectrumSolutionSet,
    SpectrumSolutionSet)


if TYPE_CHECKING:
    from glycresoft.database.mass_collection import SearchableMassCollection


def group_by_precursor_mass(scans, window_size=1.5e-5):
    scans = sorted(
        scans, key=lambda x: x.precursor_information.extracted_neutral_mass,
        reverse=True)
    groups = []
    if len(scans) == 0:
        return groups
    current_group = [scans[0]]
    last_scan = scans[0]
    for scan in scans[1:]:
        delta = abs(
            (scan.precursor_information.extracted_neutral_mass -
             last_scan.precursor_information.extracted_neutral_mass
             ) / last_scan.precursor_information.extracted_neutral_mass)
        if delta > window_size:
            groups.append(current_group)
            current_group = [scan]
        else:
            current_group.append(scan)
        last_scan = scan
    groups.append(current_group)
    return groups


class TandemClusterEvaluatorBase(TaskBase):
    tandem_cluster: List[ProcessedScan]
    structure_database: "SearchableMassCollection"
    verbose: bool
    n_processes: int
    mass_shifts: List[MassShift]
    probing_range_for_missing_precursors: int
    mass_shift_map: Mapping[str, MassShift]
    batch_size: int
    trust_precursor_fits: bool

    neutron_offset = isotopic_shift()
    solution_set_type = SpectrumSolutionSet

    def __init__(self, tandem_cluster, scorer_type, structure_database, verbose=False,
                 n_processes=1, ipc_manager=None, probing_range_for_missing_precursors=3,
                 mass_shifts=None, batch_size=DEFAULT_WORKLOAD_MAX, trust_precursor_fits=True):
        if mass_shifts is None:
            mass_shifts = [Unmodified]
        self.tandem_cluster = tandem_cluster
        self.scorer_type = scorer_type
        self.structure_database = structure_database
        self.verbose = verbose
        self.n_processes = n_processes
        self.ipc_manager = ipc_manager
        self.probing_range_for_missing_precursors = probing_range_for_missing_precursors
        self.mass_shifts = mass_shifts
        self.mass_shift_map = {
            m.name: m for m in self.mass_shifts
        }
        self.batch_size = batch_size
        self.trust_precursor_fits = trust_precursor_fits

    def _mark_hit(self, match):
        return self.structure_database.mark_hit(match)

    def _mark_batch(self):
        return self.structure_database.mark_batch()

    def search_database_for_precursors(self, mass: float, precursor_error_tolerance: float=1e-5):
        return self.structure_database.search_mass_ppm(mass, precursor_error_tolerance)

    def find_precursor_candidates(self, scan: ProcessedScan, error_tolerance: float=1e-5,
                                  probing_range: int=0, mass_shift: Optional[MassShift]=None):
        if mass_shift is None:
            mass_shift = Unmodified
        hits = []
        intact_mass = scan.precursor_information.extracted_neutral_mass
        for i in range(probing_range + 1):
            query_mass = intact_mass - (i * self.neutron_offset) - mass_shift.mass
            hits.extend(
                map(self._mark_hit,
                    self.search_database_for_precursors(query_mass, error_tolerance)))
        return hits

    def score_one(self, scan: ProcessedScan, precursor_error_tolerance=1e-5,
                  mass_shifts: List[MassShift]=None, **kwargs) -> SpectrumSolutionSet:
        """Search one MSn scan against the database and score all candidate matches

        Parameters
        ----------
        scan : ms_deisotope.ProcessedScan
            The MSn scan to search
        precursor_error_tolerance : float, optional
            The mass error tolerance for the precursor
        mass_shifts : Iterable of MassShift, optional
            The set of MassShifts to consider. Defaults to (Unmodified,)
        **kwargs
            Keyword arguments passed to :meth:`evaluate`

        Returns
        -------
        SpectrumSolutionSet
            The set of solutions for the searched scan
        """
        if mass_shifts is None:
            mass_shifts = (Unmodified,)
        solutions = []

        if (not scan.precursor_information.defaulted and self.trust_precursor_fits):
            probe = 0
        else:
            probe = self.probing_range_for_missing_precursors
        hits = []
        for mass_shift in mass_shifts:
            hits.extend(self.find_precursor_candidates(
                scan, precursor_error_tolerance, probing_range=probe,
                mass_shift=mass_shift))
            self._mark_batch()
            for structure in hits:
                result = self.evaluate(
                    scan, structure, mass_shift=mass_shift, **kwargs)
                solutions.append(result)
        out = self.solution_set_type(
            scan, solutions).sort().select_top()
        return out

    def score_all(self, precursor_error_tolerance: float=1e-5,
                  simplify: bool=False, **kwargs) -> List[SpectrumSolutionSet]:
        """Search and score all scans in :attr:`tandem_cluster`

        Parameters
        ----------
        precursor_error_tolerance : float, optional
            The mass error tolerance for the precursor
        simplify : bool, optional
            Whether or not to call :meth:`.SpectrumSolutionSet.simplify` on each
            scan's search result.
        **kwargs
            Keyword arguments passed to :meth:`evaluate`

        Returns
        -------
        list of SpectrumSolutionSet
        """
        out = []

        # loop over mass_shifts so that sequential queries deals with similar masses to
        # make better use of the database's cache
        for mass_shift in sorted(self.mass_shifts, key=lambda x: x.mass):
            # make iterable for score_one API
            mass_shift = (mass_shift,)
            for scan in self.tandem_cluster:
                solutions = self.score_one(
                    scan, precursor_error_tolerance,
                    mass_shifts=mass_shift,
                    **kwargs)
                if len(solutions) > 0:
                    out.append(solutions)
        if simplify:
            for case in out:
                case.simplify()
                case.select_top()
        return out

    def evaluate(self, scan: ProcessedScan, structure, *args, **kwargs):
        """Computes the quality of the match between ``scan``
        and the component fragments of ``structure``.

        To be overridden by implementing classes.

        Parameters
        ----------
        scan : ms_deisotope.ProcessedScan
            The scan to evaluate
        structure : object
            An object representing a structure (peptide, glycan, glycopeptide)
            to produce fragments from and search against ``scan``
        *args
            Propagated
        **kwargs
            Propagated

        Returns
        -------
        SpectrumMatcherBase
        """
        raise NotImplementedError()

    def _map_scans_to_hits(self, scans: List[ProcessedScan], precursor_error_tolerance: float=1e-5) -> WorkloadManager:
        groups = group_by_precursor_mass(
            scans, precursor_error_tolerance * 1.5)

        workload = WorkloadManager()

        i = 0
        n = len(scans)
        report_interval = 0.25 * n if n < 2000 else 0.1 * n
        last_report = report_interval
        self.log("... Begin Collecting Hits")
        for mass_shift in sorted(self.mass_shifts, key=lambda x: x.mass):
            self.log("... Mapping For %s" % (mass_shift.name,))
            i = 0
            last_report = report_interval
            for group in groups:
                if len(group) == 0:
                    continue
                i += len(group)
                report = False
                if i > last_report:
                    report = True
                    self.log(
                        "...... Mapping %0.2f%% of spectra (%d/%d) %0.4f" % (
                            i * 100. / n, i, n,
                            group[0].precursor_information.extracted_neutral_mass))
                    while last_report < i and report_interval != 0:
                        last_report += report_interval
                j = 0
                for scan in group:
                    workload.add_scan(scan)

                    # For a sufficiently dense database or large value of probe, this
                    # could easily throw the mass interval cache scheme into hysteresis.
                    # If this does occur, instead of doing this here, we could search all
                    # defaulted precursors afterwards.
                    if not scan.precursor_information.defaulted and self.trust_precursor_fits:
                        probe = 0
                    else:
                        probe = self.probing_range_for_missing_precursors
                    hits = self.find_precursor_candidates(
                        scan, precursor_error_tolerance, probing_range=probe,
                        mass_shift=mass_shift)
                    for hit in hits:
                        j += 1
                        workload.add_scan_hit(scan, hit, mass_shift.name)
                if report:
                    self.log("...... Mapping Segment Done. (%d spectrum-pairs)" % (j,))
        self._mark_batch()
        return workload

    def _get_solution_handler(self):
        if issubclass(self.solution_set_type, MultiScoreSpectrumSolutionSet):
            handler_tp = MultiScoreSolutionHandler
            score_set_type = self.scorer_type.get_score_set_type()
            packer = MultiScoreSolutionPacker(score_set_type)
            handler_tp = functools.partial(handler_tp, packer=packer)
        else:
            handler_tp = SolutionHandler
        return handler_tp

    def _transform_matched_collection(self, solution_set_collection: List[SpectrumSolutionSet],
                                      *args, **kwargs) -> List[SpectrumSolutionSet]:
        '''This helper method can be used to re-write the target
        attribute of spectrm matches. By default, it is a no-op.

        Returns
        -------
        :class:`list` of :class:`SpectrumSolutionSet`
        '''
        return solution_set_collection

    def _evaluate_hit_groups_single_process(self, workload: WorkloadManager, *args, **kwargs) -> Mapping[str, List]:
        batch_size, scan_map, hit_map, hit_to_scan_map, scan_hit_type_map, hit_group_map = workload
        handler_tp = self._get_solution_handler()
        processor = SequentialIdentificationProcessor(
            self,
            self.mass_shift_map,
            evaluation_args=kwargs,
            solution_handler_type=handler_tp)
        processor.process(scan_map, hit_map, hit_to_scan_map,
                          scan_hit_type_map, hit_group_map)
        return processor.scan_solution_map

    def collect_scan_solutions(self, scan_solution_map: Mapping[str, List],
                               scan_map: Mapping[str, ProcessedScan],
                               *args, **kwags) -> List[SpectrumSolutionSet]:
        result_set = []
        for scan_id, solutions in scan_solution_map.items():
            if len(solutions) == 0:
                continue
            scan = scan_map[scan_id]
            out = self.solution_set_type(scan, solutions).sort()
            # This is necessary to reduce the degree to which junk matches need to have their targets re-built
            out.select_top()
            if len(out) == 0:
                continue
            result_set.append(out)
        self._transform_matched_collection(result_set, *args, **kwags)
        return result_set

    @property
    def _worker_specification(self):
        raise NotImplementedError()

    def _evaluate_hit_groups_multiple_processes(self, workload: WorkloadManager, **kwargs):
        batch_size, scan_map, hit_map, hit_to_scan, scan_hit_type_map, hit_group_map = workload
        worker_type, init_args = self._worker_specification
        handler_tp = self._get_solution_handler()
        dispatcher = IdentificationProcessDispatcher(
            worker_type, self.scorer_type, evaluation_args=kwargs, init_args=init_args,
            n_processes=self.n_processes, ipc_manager=self.ipc_manager,
            mass_shift_map=self.mass_shift_map, solution_handler_type=handler_tp)
        return dispatcher.process(scan_map, hit_map, hit_to_scan, scan_hit_type_map, hit_group_map)

    def evaluate_hit_groups(self, batch, **kwargs) -> Mapping[str, List]:
        if self.n_processes == 1 or len(batch.hit_map) < 2500:
            return self._evaluate_hit_groups_single_process(
                batch, **kwargs)
        else:
            return self._evaluate_hit_groups_multiple_processes(
                batch, **kwargs)

    def score_bunch(self, scans: List[ProcessedScan],
                    precursor_error_tolerance: float=1e-5, **kwargs) -> List[SpectrumSolutionSet]:
        workload = self._map_scans_to_hits(
            scans, precursor_error_tolerance)
        solutions = []
        for batch in workload.batches(self.batch_size):
            scan_solution_map = self.evaluate_hit_groups(batch, **kwargs)
            solutions += self.collect_scan_solutions(scan_solution_map, batch.scan_map)
        return solutions



cimport cython

from ms_deisotope._c.peak_set cimport DeconvolutedPeak
from glycopeptidepy._c.structure.fragment cimport FragmentBase

ctypedef fused FragmentType:
    FragmentBase
    object


@cython.final
@cython.freelist(1000000)
cdef class PeakFragmentPair(object):

    cdef:
        public DeconvolutedPeak peak
        public object fragment
        public str fragment_name
        public Py_hash_t _hash

    @staticmethod
    cdef PeakFragmentPair _create(DeconvolutedPeak peak, fragment)
    @staticmethod
    cdef PeakFragmentPair _create_simple(DeconvolutedPeak peak, FragmentBase fragment)

    cpdef double mass_accuracy(self)


@cython.final
@cython.freelist(1000000)
cdef class PeakPairTransition(object):
    cdef:
        public DeconvolutedPeak start
        public DeconvolutedPeak end
        public object annotation
        public tuple key
        public Py_hash_t _hash

    @staticmethod
    cdef PeakPairTransition _create(DeconvolutedPeak start, DeconvolutedPeak end, object annotation)


cdef class SpectrumGraph(object):
    cdef:
        public set transitions
        public object by_first
        public object by_second

    cpdef add(self, DeconvolutedPeak p1, DeconvolutedPeak p2, object annotation)
    cpdef list topological_sort(self, adjacency_matrix=*)

    cpdef list paths_starting_at(self, ix, int limit=*)
    cpdef list paths_ending_at(self, ix, int limit=*)

    cpdef list transitive_closure(self, list paths, int limit=*)


# forward declaration
cdef class FragmentMatchMap(object)


@cython.freelist(10000)
cdef class _FragmentIndexBase(object):
    cdef:
        public FragmentMatchMap fragment_set
        public object _mapping

    cdef int _create_mapping(self)
    cdef void invalidate(self)
    cdef list getitem(self, object key)
    cdef bint has_key(self, object key)


cdef class ByFragmentIndex(_FragmentIndexBase):

    @staticmethod
    cdef ByFragmentIndex _create(FragmentMatchMap fragment_set)


cdef class ByPeakIndex(_FragmentIndexBase):

    @staticmethod
    cdef ByPeakIndex _create(FragmentMatchMap fragment_set)


cdef class ByPeakIndexIndex(_FragmentIndexBase):

    @staticmethod
    cdef ByPeakIndexIndex _create(FragmentMatchMap fragment_set)


@cython.final
@cython.freelist(1000000)
cdef class FragmentMatchMap(object):
    cdef:
        public set members
        public ByFragmentIndex by_fragment
        public ByPeakIndex by_peak
        public ByPeakIndexIndex by_peak_index

    cpdef add(self, peak, fragment=*)
    cdef void _add_direct(self, PeakFragmentPair pair)

    cpdef set fragments(self)
    cpdef FragmentMatchMap copy(self)
    cpdef FragmentMatchMap clone(self)
    cpdef clear(self)

    cpdef list peaks_for(self, fragment)
    cpdef list fragments_for(self, peak)

    cpdef invalidate(self)
    cpdef set peak_indices(self)


from operator import mul
try:
    reduce
except NameError:
    from functools import reduce

import numpy as np


def logit(x):
    return np.log(x) - np.log(1 - x)


def logitsum(xs):
    total = 0
    for x in xs:
        total += logit(x)
    return total


def prod(*x):
    return reduce(mul, x, 1)
